# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
debusine URL Configuration.

Debusine supports multi-tenancy by setting the current scope using the
``/{scope}/`` prefix at the beginning of most URLs.

This is implemented in Django by using a middleware to detect the scope from
URL, and then setting `request.urlconf` to point to the urlconf to use to
resolve URLs for that scope.

API calls use the unscoped ``/api/`` prefix in URLs. Scope for API calls is
provided using the `X-Debusine-Scope: name` HTTP header. If the header is
missing, Debusine will use the `DEBUSINE_DEFAULT_SCOPE` setting.

Unscoped service views are placed in ``/-/``

To ease transition from legacy unscoped URLs, generated urlconfs contain
best-effort redirects from the old versions of URLs to the new ones. To prevent
ambiguity, scope names are validated against a list of reserved keywords, which
contains prefixes used for non-scoped URLs.
"""

from typing import Any

from django.conf import settings
from django.contrib import admin
from django.urls import URLPattern, URLResolver, include, path, re_path
from django.views.generic import RedirectView

from debusine.web.views.auth import LoginView, LogoutView
from debusine.web.views.task_status import TaskStatusView
from debusine.web.views.views import ThemeSelectionView, ToastView

service_urlpatterns: list[URLPattern | URLResolver] = [
    path("admin/", admin.site.urls),
    path("status/queue/", TaskStatusView.as_view(), name="task-status"),
    path(
        "status/workers/",
        include("debusine.web.urls.workers", namespace="workers"),
    ),
    path(
        "status/worker-pools/",
        include("debusine.web.urls.worker_pools", namespace="worker-pools"),
    ),
    path(
        "user/",
        include("debusine.web.urls.user", namespace="user"),
    ),
    path("login/", LoginView.as_view(), name="login"),
    path("logout/", LogoutView.as_view(), name="logout"),
    path(
        "signon/",
        include("debusine.web.urls.signon", namespace="signon"),
    ),
    path(
        "enroll/",
        include("debusine.web.urls.enroll", namespace="enroll"),
    ),
    path("toasts/", ToastView.as_view(), name="test-toasts"),
    path("theme/", ThemeSelectionView.as_view(), name="theme-selection"),
]

if (
    settings.DEBUG and 'debug_toolbar' in settings.INSTALLED_APPS
):  # pragma: no cover
    import debug_toolbar

    service_urlpatterns += [
        path("__debug__/", include(debug_toolbar.urls)),
    ]

if settings.DEBUG:  # pragma: no cover
    from debusine.web.archives.urls import (
        embedded_urlpatterns as archive_urlpatterns,
    )

    service_urlpatterns += [
        path("__archives__/", include(archive_urlpatterns)),
    ]


def make_urlpatterns(scope: str) -> list[Any]:
    """
    Create the URL structure of the website given a scope name.

    This is used by ScopeMiddleware to create a different urlconf for each
    scope, to be set it in request.urlconf when the scope is known.
    """
    return [
        path('api/', include('debusine.server.urls', namespace='api')),
        path("-/", include(service_urlpatterns)),
        path("", include("debusine.web.urls.homepage", namespace="homepage")),
        re_path(
            rf"^{scope}/(?:workspace)/(?P<path>.+)",
            RedirectView.as_view(
                url=f"/{scope}/%(path)s",
                permanent=True,
                query_string=True,
            ),
        ),
        path(f"{scope}/", include('debusine.web.urls')),
        path(
            'api-auth/<path:path>',
            RedirectView.as_view(
                url="/api/auth/%(path)s",
                permanent=True,
                query_string=True,
            ),
        ),
        path(
            "accounts/oidc_callback/<name>/",
            RedirectView.as_view(
                pattern_name="signon:oidc_callback", permanent=True
            ),
        ),
        path(
            "accounts/bind_identity/<name>/",
            RedirectView.as_view(
                pattern_name="signon:bind_identity", permanent=True
            ),
        ),
        re_path(
            r"^(?P<path>(?:accounts|workspace|artifact)/.+)",
            RedirectView.as_view(
                url=f"/{settings.DEBUSINE_DEFAULT_SCOPE}/%(path)s",
                permanent=True,
                query_string=True,
            ),
        ),
        path(
            "task-status/<path:path>",
            RedirectView.as_view(
                url="/-/status/queue/%(path)s",
                permanent=True,
                query_string=True,
            ),
        ),
        path(
            "workers/<path:path>",
            RedirectView.as_view(
                url="/-/status/workers/%(path)s",
                permanent=True,
                query_string=True,
            ),
        ),
        path(
            "user/<path:path>",
            RedirectView.as_view(
                url="/-/user/%(path)s",
                permanent=True,
                query_string=True,
            ),
        ),
    ]


# Default urlpatterns
urlpatterns = make_urlpatterns(settings.DEBUSINE_DEFAULT_SCOPE)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Debusine worker controller."""

import logging
import os
import shutil
import socket
from pathlib import Path

from utils import common
from utils.server import DebusineServer
from utils.waiter import Waiter

logger = logging.getLogger(__name__)


class Worker:
    """Worker management."""

    CONFIG_DIRECTORY = Path('/etc/debusine/worker')
    TOKEN_FILE = CONFIG_DIRECTORY / 'token'
    ACTIVATION_TOKEN_FILE = CONFIG_DIRECTORY / 'activation-token'

    @classmethod
    def set_up(cls, use_activation_token: bool = True) -> None:
        """
        Set up worker and waits for the token file.

        Steps:

        - stop a possible debusine-worker
        - copy the debusine-worker's config.ini file to the appropriate
          directory
        - if use_activation_token is true:
            - create the worker in the server's database with an activation
              token
            - configure the worker using that activation token
        - start debusine-worker
        - wait for the (non-activation) token file to appear
        """
        common.run(['systemctl', 'stop', 'debusine-worker'])

        # Make sure to delete the self.CONFIG_DIRECTORY
        if cls.CONFIG_DIRECTORY.exists():
            shutil.rmtree(cls.CONFIG_DIRECTORY)
        cls.CONFIG_DIRECTORY.mkdir(parents=True)
        shutil.chown(cls.CONFIG_DIRECTORY, 'debusine-worker')

        with (
            open("examples/worker/config.ini") as orig_config_file,
            open(cls.CONFIG_DIRECTORY / "config.ini", "w") as config_file,
        ):
            for line in orig_config_file:
                if line.startswith("api-url = "):
                    config_file.write(
                        f"api-url = https://{socket.getfqdn()}/api\n"
                    )
                else:
                    config_file.write(line)

        if use_activation_token:
            created_worker = DebusineServer.execute_command(
                "worker", "create", socket.getfqdn()
            )
            if created_worker.returncode != 0:
                raise AssertionError(
                    f"'debusine-admin worker create' exited with code"
                    f" {created_worker.returncode};"
                    f" stderr: {created_worker.stderr}"
                )
            if not created_worker.stdout:
                raise AssertionError(
                    "'debusine-admin worker create' did not print a worker"
                    " activation token"
                )

            with open(cls.ACTIVATION_TOKEN_FILE, "w") as token_file:
                os.fchmod(token_file.fileno(), 0o600)
                token_file.write(created_worker.stdout)
            shutil.chown(cls.ACTIVATION_TOKEN_FILE, "debusine-worker")

        common.run(['systemctl', 'start', 'debusine-worker'])

        cls.wait_for_token_file()

    @classmethod
    def wait_for_token_file(cls) -> bool:
        """
        Wait for the /etc/debusine/worker/token file to appear.

        :return: True if the file appeared, False if it timed out (30 seconds)
        """
        return Waiter.wait_for_success(30, cls.TOKEN_FILE.exists)

    @classmethod
    def read_token(cls) -> str:
        """Return token: contents of cls.TOKEN_FILE."""
        with open(cls.TOKEN_FILE) as token_file:
            return token_file.read()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the collection relations views."""

import re
from typing import Any, ClassVar, override
from urllib.parse import urlencode

from django.urls import reverse
from rest_framework import status
from rest_framework.settings import api_settings

from debusine.artifacts.models import CollectionCategory
from debusine.client.models import (
    CollectionIdentifierRequest,
    CollectionIdentifierResponse,
    CollectionRelationSetResponse,
    CollectionRelationType,
)
from debusine.db.models import Collection, CollectionRelation, User, Workspace
from debusine.db.playground import scenarios
from debusine.server.views.tests.base import TestCase


class CollectionRelationListViewTests(TestCase):
    """Tests for CollectionRelationListView."""

    scenario = scenarios.DefaultContextAPI()

    sid: ClassVar[Collection]
    trixie: ClassVar[Collection]
    experimental: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.sid = cls.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        cls.trixie = cls.playground.create_collection(
            "trixie", CollectionCategory.SUITE
        )
        cls.experimental = cls.playground.create_collection(
            "experimental", CollectionCategory.SUITE
        )

    def list_url(self) -> str:
        return reverse(
            "api:collection-relation-list",
            kwargs={"workspace": self.scenario.workspace.name},
        )

    def assert_list(
        self, api_response: dict[str, Any], expected: list[dict[str, Any]]
    ) -> None:
        self.assertEqual(api_response["count"], len(expected))
        self.assertEqual(api_response["results"], expected)

    def test_list_relations(self) -> None:
        """The list API returns ordered relation records."""
        CollectionRelation.objects.create(
            source=self.sid,
            target=self.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=1,
            created_by_user=self.scenario.user,
        )
        CollectionRelation.objects.create(
            source=self.sid,
            target=self.experimental,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=2,
            created_by_user=self.scenario.user,
        )

        response = self.client.get(
            self.list_url(), headers={"Token": self.scenario.user_token.key}
        )
        data = self.assertAPIResponseOk(response)

        self.assert_list(
            data,
            [
                {
                    "source_collection": {
                        "category": CollectionCategory.SUITE,
                        "name": "sid",
                    },
                    "target_collection": {
                        "category": CollectionCategory.SUITE,
                        "name": "trixie",
                    },
                    "position": 1,
                    "type": CollectionRelation.Types.SUITE_REQUIRES,
                },
                {
                    "source_collection": {
                        "category": CollectionCategory.SUITE,
                        "name": "sid",
                    },
                    "target_collection": {
                        "category": CollectionCategory.SUITE,
                        "name": "experimental",
                    },
                    "position": 2,
                    "type": CollectionRelation.Types.SUITE_REQUIRES,
                },
            ],
        )

    def test_list_relations_target_filter(self) -> None:
        CollectionRelation.objects.create(
            source=self.sid,
            target=self.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=1,
            created_by_user=self.scenario.user,
        )
        CollectionRelation.objects.create(
            source=self.sid,
            target=self.experimental,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=2,
            created_by_user=self.scenario.user,
        )

        response = self.client.get(
            self.list_url() + "?target=trixie@debian:suite",
            headers={"Token": self.scenario.user_token.key},
        )
        data = self.assertAPIResponseOk(response)

        self.assert_list(
            data,
            [
                {
                    "source_collection": {
                        "category": CollectionCategory.SUITE,
                        "name": "sid",
                    },
                    "target_collection": {
                        "category": CollectionCategory.SUITE,
                        "name": "trixie",
                    },
                    "position": 1,
                    "type": CollectionRelation.Types.SUITE_REQUIRES,
                }
            ],
        )

    def test_list_relations_etag(self) -> None:
        response = self.client.get(
            self.list_url() + "?source=sid@debian:suite&type=suite-based-on",
            headers={"Token": self.scenario.user_token.key},
        )

        etag = response["ETag"]

        self.assertIsNotNone(etag)

        CollectionRelation.objects.create(
            source=self.sid,
            target=self.trixie,
            type=CollectionRelation.Types.SUITE_BASED_ON,
            created_by_user=self.scenario.user,
        )

        response = self.client.get(
            self.list_url() + "?source=sid@debian:suite&type=suite-based-on",
            headers={"Token": self.scenario.user_token.key},
        )

        self.assertNotEqual(response["ETag"], etag)

    def test_list_relations_source_type_no_pagination_yes_etag(self) -> None:
        page_size = api_settings.PAGE_SIZE
        assert page_size is not None

        for number in range(page_size + 1):
            suite = self.playground.create_collection(
                f"suite-{number}", CollectionCategory.SUITE
            )
            CollectionRelation.objects.create(
                source=self.trixie,
                target=suite,
                type=CollectionRelation.Types.SUITE_REQUIRES,
                position=number,
                created_by_user=self.scenario.user,
            )

        response = self.client.get(
            self.list_url() + "?source=trixie@debian:suite&type=suite-requires",
            headers={"Token": self.scenario.user_token.key},
        )

        self.assertIn("ETag", response.headers)

        data = self.assertAPIResponseOk(response)

        relations = page_size + 1

        self.assertEqual(len(data["results"]), relations)

    def test_list_relations_yes_pagination_no_etag(self) -> None:
        page_size = api_settings.PAGE_SIZE
        assert page_size is not None

        for number in range(page_size + 1):
            suite = self.playground.create_collection(
                f"suite-{number}", CollectionCategory.SUITE
            )
            CollectionRelation.objects.create(
                source=self.trixie,
                target=suite,
                type=CollectionRelation.Types.SUITE_REQUIRES,
                position=number,
                created_by_user=self.scenario.user,
            )

        response = self.client.get(
            self.list_url() + "?source=trixie@debian:suite",
            headers={"Token": self.scenario.user_token.key},
        )

        self.assertNotIn("ETag", response.headers)

        data = self.assertAPIResponseOk(response)

        self.assertEqual(len(data["results"]), page_size)
        self.assertEqual(data["count"], page_size + 1)

    def test_list_relations_source_filter(self) -> None:
        CollectionRelation.objects.create(
            source=self.sid,
            target=self.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=1,
            created_by_user=self.scenario.user,
        )
        CollectionRelation.objects.create(
            source=self.trixie,
            target=self.experimental,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=2,
            created_by_user=self.scenario.user,
        )

        response = self.client.get(
            self.list_url() + "?source=sid@debian:suite",
            headers={"Token": self.scenario.user_token.key},
        )
        data = self.assertAPIResponseOk(response)

        self.assert_list(
            data,
            [
                {
                    "source_collection": {
                        "category": CollectionCategory.SUITE,
                        "name": "sid",
                    },
                    "target_collection": {
                        "category": CollectionCategory.SUITE,
                        "name": "trixie",
                    },
                    "position": 1,
                    "type": CollectionRelation.Types.SUITE_REQUIRES,
                }
            ],
        )

    def test_list_relations_type_filter(self) -> None:
        CollectionRelation.objects.create(
            source=self.sid,
            target=self.trixie,
            type=CollectionRelation.Types.SUITE_BASED_ON,
            created_by_user=self.scenario.user,
        )
        CollectionRelation.objects.create(
            source=self.trixie,
            target=self.experimental,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=2,
            created_by_user=self.scenario.user,
        )

        response = self.client.get(
            self.list_url() + "?type=suite-based-on",
            headers={"Token": self.scenario.user_token.key},
        )
        data = self.assertAPIResponseOk(response)

        self.assert_list(
            data,
            [
                {
                    "source_collection": {
                        "category": CollectionCategory.SUITE,
                        "name": "sid",
                    },
                    "target_collection": {
                        "category": CollectionCategory.SUITE,
                        "name": "trixie",
                    },
                    "position": None,
                    "type": CollectionRelation.Types.SUITE_BASED_ON,
                }
            ],
        )

    def test_list_relations_invalid_type_filter(self) -> None:
        response = self.client.get(
            self.list_url() + "?type=something-invalid",
            headers={"Token": self.scenario.user_token.key},
        )
        self.assertResponseProblem(
            response,
            "Invalid relation type",
            detail_pattern="Unknown relation type: 'something-invalid'",
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    def test_list_relation_invalid_collection_identifier(self) -> None:
        collection = "does-not/exist@debian:suite"

        response = self.client.get(
            self.list_url() + f"?source={collection}",
            headers={"Token": self.scenario.user_token.key},
        )

        self.assertResponseProblem(
            response,
            "Invalid collection identifier in query parameter",
            detail_pattern=(
                f"Collection identifier {collection!r} must not contain '/'"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    def test_list_relation_collection_filter_does_not_exist(self) -> None:
        collection = "does-not-exist@debian:suite"

        response = self.client.get(
            self.list_url() + f"?source={collection}",
            headers={"Token": self.scenario.user_token.key},
        )

        self.assertResponseProblem(
            response,
            "Collection not found",
            detail_pattern=(
                f"Collection '{collection}' does not "
                f"exist in workspace 'System'"
            ),
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_list_relations_invalid_source_format_query_parameter(self) -> None:
        response = self.client.get(
            self.list_url() + "?source=sid",
            headers={"Token": self.scenario.user_token.key},
        )

        self.assertResponseProblem(
            response,
            "Invalid collection identifier in query parameter",
            detail_pattern="sid",
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    def test_list_relations_invalid_source_category_query_parameter(
        self,
    ) -> None:
        response = self.client.get(
            self.list_url() + "?source=sid@invalid-category",
            headers={"Token": self.scenario.user_token.key},
        )

        self.assertResponseProblem(
            response,
            "Collection not found",
            detail_pattern=(
                "Collection 'sid@invalid-category' "
                "does not exist in workspace 'System'"
            ),
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_unauthenticated(self) -> None:
        """Authentication is required."""
        response = self.client.get(self.list_url())

        self.assertResponseProblem(
            response,
            "Error",
            detail_pattern="Authentication credentials were not provided.",
            status_code=status.HTTP_403_FORBIDDEN,
        )


class CollectionRelationSetViewTests(TestCase):
    """Tests for CollectionRelationSetView."""

    scenario = scenarios.DefaultContextAPI()

    trixie: ClassVar[Collection]
    trixie_backports: ClassVar[Collection]
    personal: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.trixie = cls.playground.create_collection(
            "trixie", CollectionCategory.SUITE
        )
        cls.trixie_backports = cls.playground.create_collection(
            "trixie-backports", CollectionCategory.SUITE
        )
        cls.personal = cls.playground.create_collection(
            "personal", CollectionCategory.SUITE
        )

    def grant_user_can_create_relations(
        self, user: User, collection: Collection
    ) -> None:
        owners = self.playground.create_group_role(
            collection.workspace, Workspace.Roles.OWNER
        )
        self.playground.add_user(owners, user)

    def set_url(self, *, relation_type: CollectionRelation.Types) -> str:
        return reverse(
            "api:collection-relation-update",
            kwargs={
                "workspace": self.scenario.workspace.name,
                "collection": "personal@debian:suite",
                "relation_type": relation_type,
            },
        )

    def get_etag(self, *, relation_type: CollectionRelation.Types) -> str:
        url = reverse(
            "api:collection-relation-list",
            kwargs={"workspace": self.scenario.workspace.name},
        )

        url += f"?source=personal@debian:suite&type={relation_type}"

        response = self.client.get(
            url, headers={"Token": self.scenario.user_token.key}
        )

        return response.headers["ETag"]

    def test_set_relationships(self) -> None:
        user = self.scenario.user

        self.grant_user_can_create_relations(user, self.personal)

        targets = [
            CollectionIdentifierRequest(
                name="trixie", category=CollectionCategory.SUITE
            ),
            CollectionIdentifierRequest(
                name="trixie-backports", category=CollectionCategory.SUITE
            ),
        ]

        self.assertEqual(
            CollectionRelation.objects.filter(source=self.personal).count(), 0
        )

        etag = self.get_etag(
            relation_type=CollectionRelation.Types.SUITE_REQUIRES
        )

        response = self.get_debusine().collection_relation_set(
            self.scenario.workspace.name,
            source="personal@debian:suite",
            relation_type=CollectionRelationType.SUITE_REQUIRES,
            targets=targets,
            if_match=etag,
        )

        self.assertEqual(
            response,
            CollectionRelationSetResponse(
                targets=[
                    CollectionIdentifierResponse(
                        category=CollectionCategory.SUITE,
                        name="trixie",
                    ),
                    CollectionIdentifierResponse(
                        category=CollectionCategory.SUITE,
                        name="trixie-backports",
                    ),
                ]
            ),
        )

        trixie, trixie_backports = CollectionRelation.objects.filter(
            source=self.personal
        ).order_by("position")

        self.assertEqual(trixie.target, self.trixie)
        self.assertEqual(trixie_backports.target, self.trixie_backports)

    def test_set_relationships_failed_invalid_relation_targets(self) -> None:
        user = self.scenario.user

        self.grant_user_can_create_relations(user, self.personal)

        etag = self.get_etag(
            relation_type=CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS
        )

        response = self.client.put(
            self.set_url(
                relation_type=CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS
            ),
            headers={"Token": self.scenario.user_token.key, "If-Match": etag},
            data={
                "targets": [
                    {"name": "trixie", "category": CollectionCategory.SUITE},
                ]
            },
            content_type="application/json",
        )

        self.assertResponseProblem(
            response,
            title="Invalid relation targets",
            detail_pattern="db_collectionrelation_categories_correct",
        )

    def test_set_relationships_failed_no_if_match(self) -> None:
        user = self.scenario.user

        self.grant_user_can_create_relations(user, self.personal)

        response = self.client.put(
            self.set_url(relation_type=CollectionRelation.Types.SUITE_REQUIRES),
            headers={"Token": self.scenario.user_token.key},
            data={
                "targets": [
                    {"name": "trixie", "category": CollectionCategory.SUITE},
                ]
            },
            content_type="application/json",
        )

        self.assertResponseProblem(
            response,
            title='"If-Match" header is required',
            detail_pattern=(
                'The "If-Match" header is required for PUT. '
                'Fetch the current ETag from'
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    def test_set_relationships_failed_invalid_if_match(self) -> None:
        user = self.scenario.user

        self.grant_user_can_create_relations(user, self.personal)

        etag = self.get_etag(
            relation_type=CollectionRelation.Types.SUITE_REQUIRES
        )

        CollectionRelation.objects.create(
            source=self.personal,
            target=self.trixie_backports,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=1,
            created_by_user=self.scenario.user,
        )

        response = self.client.put(
            self.set_url(relation_type=CollectionRelation.Types.SUITE_REQUIRES),
            headers={"Token": self.scenario.user_token.key, "If-Match": etag},
            data={
                "targets": [
                    {"name": "trixie", "category": CollectionCategory.SUITE},
                ]
            },
            content_type="application/json",
        )

        list_url = (
            reverse(
                "api:collection-relation-list",
                kwargs={"workspace": self.scenario.workspace.name},
            )
            + "?"
            + urlencode(
                {
                    "source": "personal@debian:suite",
                    "type": "suite-requires",
                }
            )
        )
        self.assertResponseProblem(
            response,
            "Collection relations changed",
            detail_pattern=re.escape(
                'The "If-Match" header does not match the '
                f'current ETag. Fetch the current ETag from '
                f'"{list_url}" and retry.'
            ),
            status_code=status.HTTP_412_PRECONDITION_FAILED,
        )

    def test_set_relationship_target_collection_does_not_exist(self) -> None:
        user = self.scenario.user

        self.grant_user_can_create_relations(user, self.personal)

        etag = self.get_etag(
            relation_type=CollectionRelation.Types.SUITE_REQUIRES
        )

        response = self.client.put(
            self.set_url(relation_type=CollectionRelation.Types.SUITE_REQUIRES),
            headers={"Token": self.scenario.user_token.key, "If-Match": etag},
            data={
                "targets": [
                    {
                        "name": "does-not-exist",
                        "category": CollectionCategory.SUITE,
                    },
                ]
            },
            content_type="application/json",
        )

        self.assertResponseProblem(
            response,
            "Collection 'does-not-exist@debian:suite' "
            "does not exist or is not accessible",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Infrastructure to create mock artifact data."""

import datetime as dt
import email.utils
import hashlib
import re
import subprocess
import tempfile
import textwrap
from collections.abc import Set
from pathlib import Path, PurePath
from typing import Any

import debian.deb822 as deb822

from debusine.artifacts.local_artifact import (
    BinaryPackage,
    PackageBuildLog,
    SourcePackage,
    Upload,
    deb822dict_to_dict,
)


class ArtifactPlayground:
    """Create mock artifact data."""

    @classmethod
    def create_source_package(
        cls,
        destdir: Path,
        *,
        name: str = "hello",
        binaries: list[str] | None = None,
        version: str = "1.0-1",
        architectures: set[str] | None = None,
    ) -> SourcePackage:
        """Create a SourcePackage artifact."""
        dsc = destdir / f"{name}_{version}.dsc"
        sources: list[Path] = []
        v_parts = version.split("-", 1)
        if len(v_parts) == 2:
            sources.append(destdir / f"{name}_{version}.debian.tar.xz")
            sources.append(destdir / f"{name}_{v_parts[0]}.orig.tar.gz")
        else:
            sources.append(destdir / f"{name}_{version}.tar.gz")
        for source in sources:
            source.write_bytes(source.name.encode())
        cls.write_deb822_file(
            deb822.Dsc,
            dsc,
            sources,
            source=name,
            binaries=binaries,
            version=version,
            architectures=architectures,
        )
        return SourcePackage.create(
            name=name, version=version, files=[dsc, *sources]
        )

    @classmethod
    def create_binary_package(
        cls,
        destdir: Path,
        *,
        name: str = "hello",
        version: str = "1.0-1",
        architecture: str = "all",
        source_name: str | None = None,
        source_version: str | None = None,
    ) -> BinaryPackage:
        """Create a BinaryPackage artifact."""
        deb = destdir / f"{name}_{version}_{architecture}.deb"
        cls.write_deb_file(
            deb,
            source_name=source_name or name,
            source_version=source_version or version,
        )
        return BinaryPackage.create(file=deb)

    @classmethod
    def create_upload(
        cls,
        destdir: Path,
        *,
        source: bool = True,
        src_name: str = "hello",
        version: str = "1.0-1",
        src_architectures: set[str] | None = None,
        binary: bool = False,
        binaries: list[tuple[str, str]] | None = None,
        section: str = "unknown",
        priority: str = "optional",
    ) -> Upload:
        """Create an Upload artifact."""
        architectures: set[str] = set()
        files: list[Path] = []
        if binaries is None:
            binaries = [(src_name, "all")]
        if source:
            architectures.add("source")
            dsc = destdir / f"{src_name}_{version}.dsc"
            src_files: list[Path] = []
            v_parts = version.split("-", 1)
            if len(v_parts) == 2:
                src_files.append(
                    destdir / f"{src_name}_{version}.debian.tar.xz"
                )
                src_files.append(
                    destdir / f"{src_name}_{v_parts[0]}.orig.tar.gz"
                )
            else:
                src_files.append(destdir / f"{src_name}_{version}.tar.gz")
            for src_file in src_files:
                src_file.write_bytes(src_file.name.encode())
            cls.write_deb822_file(
                deb822.Dsc,
                dsc,
                src_files,
                source=src_name,
                binaries=[name for (name, architecture) in binaries],
                version=version,
                architectures=src_architectures,
            )
            files += [dsc, *src_files]
        if binary:
            for bin_name, bin_architecture in binaries:
                architectures.add(bin_architecture)
                deb = destdir / f"{bin_name}_{version}_{bin_architecture}.deb"
                cls.write_deb_file(
                    deb,
                    source_name=src_name,
                    source_version=version,
                    extra_control_fields={
                        "Section": section,
                        "Priority": priority,
                    },
                )
                files.append(deb)
        else:
            binaries = None
        arch_tag = "+".join(sorted(architectures))
        changes = destdir / f"{src_name}_{version}_{arch_tag}.changes"
        cls.write_deb822_file(
            deb822.Changes,
            changes,
            files,
            source=src_name,
            binaries=(
                [name for (name, architecture) in binaries]
                if binaries
                else None
            ),
            version=version,
            section=section,
        )
        return Upload.create(changes_file=changes)

    @classmethod
    def create_package_build_log(
        cls,
        file: Path,
        *,
        source: str | None = None,
        version: str | None = None,
        architecture: str | None = None,
    ) -> PackageBuildLog:
        """Create a PackageBuildLog artifact."""
        if source is None or version is None:
            if m := re.match(
                r"([a-z][a-z0-9+-.]+)"
                r"_([0-9][A-Za-z0-9+-.~:]*)"
                r"_([a-z0-9-]+).build$",
                file.name,
            ):
                if source is None:
                    source = m.group(1)
                if version is None:
                    version = m.group(2)
            else:
                raise ValueError(
                    "source or version not provided"
                    " and cannot be inferred from file name"
                )

        return PackageBuildLog.create(
            file=file,
            source=source,
            version=version,
            architecture=architecture,
        )

    @classmethod
    def write_deb_file(
        cls,
        path: Path,
        *,
        source_name: str | None = None,
        source_version: str | None = None,
        control_file_names: list[str] | None = None,
        maintainer: str = "Example Maintainer <example@example.org>",
        extra_control_fields: dict[str, str] | None = None,
        data_files: dict[PurePath, bytes] | None = None,
        data_symlinks: dict[PurePath, PurePath] | None = None,
    ) -> dict[str, Any]:
        """Write a debian control file."""
        with tempfile.TemporaryDirectory() as tempdir:
            build_directory = Path(tempdir)
            (build_directory / "DEBIAN").mkdir()
            if m := re.match(
                r"([a-z][a-z0-9+-.]+)"
                r"_([0-9][A-Za-z0-9+-.~:]*)"
                r"_([a-z0-9-]+).u?deb$",
                path.name,
            ):
                package, version, architecture = m.groups()
            else:  # pragma: no cover
                raise ValueError(f"Badly-formed .deb file name: {path}")
            source = source_name if source_name is not None else package
            if source_version is not None and source_version != version:
                source += f" ({source_version})"

            control_contents = textwrap.dedent(
                f"""\
                Package: {package}
                Version: {version}
                Architecture: {architecture}
                Maintainer: {maintainer}
                Description: Example description
                """
            )
            if source != package:
                control_contents += f"Source: {source}\n"
            if extra_control_fields is not None:
                for key, value in extra_control_fields.items():
                    control_contents += f"{key}: {value}\n"
            (build_directory / "DEBIAN" / "control").write_text(
                control_contents
            )
            for name in control_file_names or []:
                (build_directory / "DEBIAN" / name).touch()
            if data_files is not None:
                for data_path, data_contents in data_files.items():
                    assert not data_path.is_absolute()
                    (build_directory / data_path).parent.mkdir(
                        parents=True, exist_ok=True
                    )
                    (build_directory / data_path).write_bytes(data_contents)
            if data_symlinks is not None:
                for data_path, target in data_symlinks.items():
                    assert not data_path.is_absolute()
                    (build_directory / data_path).parent.mkdir(
                        parents=True, exist_ok=True
                    )
                    (build_directory / data_path).symlink_to(target)
            subprocess.run(
                [
                    "dpkg-deb",
                    "--root-owner-group",
                    "--build",
                    build_directory,
                    path,
                ],
                stdout=subprocess.DEVNULL,
            )

            return deb822dict_to_dict(deb822.Deb822(control_contents.encode()))

    @classmethod
    def write_deb822_file(
        cls,
        file_type: type[deb822.Deb822],
        path: Path,
        files: list[Path],
        *,
        format_version: str | None = None,
        source: str = "hello-traditional",
        binaries: list[str] | None = None,
        version: str,
        architectures: set[str] | None = None,
        binnmu: bool = False,
        maintainer: str = "Example Maintainer <example@example.org>",
        section: str = "unknown",
    ) -> dict[str, Any]:
        """Write dsc or changes file with files information."""
        architectures = set(architectures or set())
        match file_type:
            case deb822.Dsc:
                architectures.add("any")
                format_version = format_version or "3.0 (quilt)"
            case deb822.Changes:
                for file in files:
                    name = file.name
                    if name.endswith(".dsc"):
                        architectures.add("source")
                    elif name.endswith((".deb", ".udeb")):
                        if m := re.match(r".*_([a-z0-9-]+).u?deb$", name):
                            architectures.add(m.group(1))
                format_version = format_version or "1.8"
            case _ as unreachable:
                raise NotImplementedError(
                    f"{unreachable!r} not supported:"
                    " only Dsc and Changes are supported"
                )

        if binnmu:
            source += f" ({version})"
            version += "+b1"

        changes_contents = textwrap.dedent(
            f"""\
            Format: {format_version}
            Source: {source}
            Architecture: {' '.join(sorted(architectures))}
            Version: {version}
            Maintainer: {maintainer}
            """
        )
        match file_type:
            case deb822.Dsc:
                if binaries is None:
                    binaries = [source]
                changes_contents += textwrap.dedent(
                    f"""\
                    Binary: {', '.join(binaries)}
                    Homepage: http://www.gnu.org/software/{source}/
                    Standards-Version: 4.3.0
                    Package-List:
                    """
                ) + "".join(
                    f" {binary} deb devel optional arch=any\n"
                    for binary in binaries
                )
            case deb822.Changes:
                timestamp = dt.datetime(
                    2024, 12, 1, 0, 0, 0, tzinfo=dt.UTC
                ).timestamp()
                changes_contents += textwrap.dedent(
                    f"""\
                    Date: {email.utils.formatdate(timestamp)}
                    Distribution: unstable
                    Urgency: medium
                    Changes:
                      {source} ({version}) unstable; urgency=medium
                      .
                        * Test upload.
                    """
                )
                if binaries:
                    changes_contents += (
                        f"Binary: {' '.join(binaries)}\n"
                        + "Description:\n"
                        + "\n".join(
                            f" {binary} - A Description" for binary in binaries
                        )
                        + "\n"
                    )
            case _ as unreachable:
                raise NotImplementedError(
                    f"{unreachable!r} not supported:"
                    " only Dsc and Changes are supported"
                )
        changes_contents += cls.hash_deb822_files(
            file_type,
            {file.name: file.read_bytes() for file in files},
            section=section,
        )
        path.write_text(changes_contents)

        return deb822dict_to_dict(file_type(path.read_bytes()))

    @staticmethod
    def hash_deb822_files(
        file_type: type[deb822.Deb822],
        file_contents: dict[str, bytes],
        break_hashes: Set[str] = frozenset(),
        break_sizes: Set[str] = frozenset(),
        section: str = "unknown",
    ) -> str:
        """
        Build the file-manifest parts of a Deb822 file that we're generating.

        Hash the files in file_contents, a dict of {filename: file_contents}.
        file_type refers to the kind of Deb822 object that we're generating.
        For any hash named in break_hashes, the relevant hashes will be
        incorrect. break_sizes works similarly, but alters the size of the
        file, in the named hash block.

        The returned string is a block of deb822-formatted file content.
        """
        output = []
        for hash_ in ("sha256", "sha1", "md5"):
            if hash_ == "md5":
                output.append("Files:")
            else:
                output.append(f"Checksums-{hash_.title()}:")
            for filename, contents in file_contents.items():
                hasher = hashlib.new(hash_)
                hasher.update(contents)
                size = len(contents)
                if hash_ in break_hashes:
                    hasher.update(b"oops!")
                if hash_ in break_sizes:
                    size += 1
                if hash_ == "md5" and file_type == deb822.Changes:
                    output.append(
                        f" {hasher.hexdigest()} {size} {section} optional "
                        f"{filename}"
                    )
                else:
                    output.append(f" {hasher.hexdigest()} {size} {filename}")
        return "\n".join(output)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine command line interface, workspace management commands."""

import argparse
import sys
from typing import Any, override

from debusine.client.commands.base import WorkspaceCommand
from debusine.client.exceptions import DebusineError
from debusine.client.models import WorkspaceInheritanceChain
from debusine.utils.input import YamlEditor


class Inheritance(WorkspaceCommand, group="workspace"):
    """Query and change the inheritance chain of a workspace."""

    def __init__(
        self, parser: argparse.ArgumentParser, args: argparse.Namespace
    ) -> None:
        """Handle workspace also as a (deprecated) positional argument."""
        super().__init__(parser, args)
        # Override with the workspace given as positional argument, if present
        if self.args.workspace_arg:
            self.workspace = self.args.workspace_arg

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument("workspace_arg", nargs="?", help="workspace name")
        parser.add_argument(
            "--append",
            type=str,
            nargs="+",
            help="workspaces (id, 'name', or 'scope/name')"
            " to add to the end of the parent list",
        )
        parser.add_argument(
            "--prepend",
            type=str,
            nargs="+",
            help="workspaces (id, 'name', or 'scope/name')"
            " to add to the beginning of the parent list",
        )
        parser.add_argument(
            "--remove",
            type=str,
            nargs="+",
            help="workspaces (id, 'name', or 'scope/name')"
            " to remove from the parent list",
        )
        parser.add_argument(
            "--set",
            type=str,
            nargs="+",
            help="workspaces (id, 'name', or 'scope/name')"
            " that compose the new parent list",
        )
        parser.add_argument(
            "--edit",
            action="store_true",
            help="edit the parent list using $EDITOR",
        )

    def _apply_inheritance(
        self,
        *,
        workspace: str,
        chain: WorkspaceInheritanceChain,
    ) -> WorkspaceInheritanceChain:
        """
        Set inheritance.

        If API call fails because of an error not caused by the parameters
         fail.

        :raises DebusineError: if an error setting the inheritance happens
          (e.g. invalid chain because workspace does not exist).
        """
        with self._api_call_or_fail():
            return self.debusine.set_workspace_inheritance(
                workspace, chain=chain
            )

    def _edit_and_apply_inheritance(
        self,
        workspace: str,
        current: WorkspaceInheritanceChain,
    ) -> WorkspaceInheritanceChain | None:
        """Run the interactive YAML edit flow and return new inheritance."""
        editor: YamlEditor[list[Any]] = YamlEditor(
            current.model_dump()["chain"],
            help_text=[
                "Here you can configure the inheritance"
                f" of workspace '{workspace}'.",
                "",
                "This is a YAML list of dictionaries with"
                " 'id/scope/workspace' keys.",
                "Usually you specify the workspace,"
                " and optionally the scope. The scope",
                "is only required to refer to a workspace"
                " outside of the current scope.",
                "",
                "For example:",
                "",
                "- workspace: base",
                "- scope: some-other-scope",
                "  workspace: foobar",
            ],
        )

        return editor.edit_apply_loop(
            apply=lambda new: self._apply_inheritance(
                workspace=workspace, chain=WorkspaceInheritanceChain(chain=new)
            ),
            apply_exception_class=DebusineError,
            print_exception=self.print_server_error,
            content_description="workspace inheritance",
        )

    @override
    def run(self) -> None:
        workspace = self.workspace
        with self._api_call_or_fail():
            try:
                current = self.debusine.get_workspace_inheritance(workspace)
            except DebusineError as err:
                output = {"result": "failure", "error": err.asdict()}
                self._print_yaml(output, file=sys.stderr)
                raise SystemExit(3)

        new = current
        if self.args.append:
            new = new + WorkspaceInheritanceChain.from_strings(self.args.append)
        if self.args.prepend:
            new = (
                WorkspaceInheritanceChain.from_strings(self.args.prepend) + new
            )
        if self.args.remove:
            new = new - WorkspaceInheritanceChain.from_strings(self.args.remove)
        if self.args.set:
            new = WorkspaceInheritanceChain.from_strings(self.args.set)
        if self.args.edit:
            result_applied = self._edit_and_apply_inheritance(
                workspace, current
            )
            to_print = result_applied if result_applied is not None else current
            self._print_yaml(to_print.model_dump())
            return

        if new != current:
            try:
                new = self._apply_inheritance(workspace=workspace, chain=new)
            except DebusineError as error:
                print("Error setting inheritance:\n", file=sys.stderr)
                self.print_server_error(error)
                raise SystemExit(3)

        self._print_yaml(new.model_dump())


class LegacyInheritance(
    Inheritance,
    name="workspace-inheritance",
    deprecated="see `workspace inheritance`",
):
    """Query and change the inheritance chain of a workspace."""



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""The incus executor backend."""

import importlib.resources
import logging
import os
import random
import string
import subprocess
import tarfile
import textwrap
import time as tm
from abc import abstractmethod
from io import BytesIO
from pathlib import Path, PurePath
from subprocess import (
    CalledProcessError,
    CompletedProcess,
    DEVNULL,
    check_call,
    check_output,
)
from tarfile import TarFile, TarInfo
from tempfile import TemporaryDirectory
from typing import Any, AnyStr, ClassVar, Literal, overload, override

import yaml
from prometheus_client.parser import text_string_to_metric_families

from debusine import utils
from debusine.client.debusine import Debusine
from debusine.client.models import ArtifactResponse
from debusine.tasks.executors.base import (
    ExecutorImageCategory,
    ExecutorInterface,
    ExecutorStatistics,
    ImageImportError,
    InstanceInterface,
    InstanceNotRunning,
)
from debusine.tasks.executors.images import ImageCache
from debusine.tasks.models import BackendType

# https://linuxcontainers.org/incus/docs/main/architectures/
DEB_TO_INCUS_ARCH = {
    "amd64": "x86_64",
    "arm64": "aarch64",
    "armel": "armv6l",
    "armhf": "armv7l",
    "i386": "i686",
    "loong64": "loongarch64",
    "mips64el": "mips64",
    "mipsel": "mips",
    "powerpc": "ppc",
    "ppc64": "ppc64",
    "ppc64el": "ppc64le",
    "riscv64": "riscv64",
    "s390x": "s390x",
}

TEMPLATES = {
    "hosts.tpl": textwrap.dedent(
        """\
        127.0.1.1   {{ instance.name }}
        127.0.0.1   localhost
        ::1         localhost ip6-localhost ip6-loopback
        ff02::1     ip6-allnodes
        ff02::2     ip6-allrouters
        """
    ),
    "hostname.tpl": "{{ instance.name }}",
}

log = logging.getLogger(__name__)


def run_incus_cmd(*args: str, suppress_stderr: bool = False) -> str:
    """Execute an incus CLI command, return the output."""
    cmd = ["incus"]
    cmd.extend(args)
    log.debug("Executing %r", cmd)
    return check_output(
        cmd,
        text=True,
        encoding="utf-8",
        stderr=DEVNULL if suppress_stderr else None,
    )


class IncusImageCache(ImageCache):
    """
    Extensions to ImageCache to import images into Incus.

    Adds hooks for adding files to the import tarball, expecting to be
    subclassed for LXC and VM details.
    """

    @override
    def image_in_cache(self, artifact: ArtifactResponse) -> bool:
        try:
            run_incus_cmd(
                "image", "info", f"artifact/{artifact.id}", suppress_stderr=True
            )
        except CalledProcessError:
            return False
        return True

    @override
    def process_downloaded_image(self, artifact: ArtifactResponse) -> None:
        src_path = self.cache_artifact_image_path(
            artifact.id, artifact.data["filename"]
        )
        with TemporaryDirectory(prefix="debusine-incus-image-build-") as wd:
            dest_path = self.build_incus_image(src_path, Path(wd), artifact)
            log.info(
                "Importing %s %i into Incus", artifact.category, artifact.id
            )
            run_incus_cmd(
                "image",
                "import",
                str(dest_path),
                "--alias",
                f"artifact/{artifact.id}",
            )

    def _image_metadata(self, artifact: ArtifactResponse) -> dict[str, Any]:
        """Return the metadata for the given artifact."""
        architecture = artifact.data["architecture"]
        vendor = artifact.data["vendor"]
        codename = artifact.data["codename"]
        return {
            "architecture": DEB_TO_INCUS_ARCH.get(architecture, architecture),
            "creation_date": int(artifact.created_at.timestamp()),
            "profiles": ["debusine"],
            "properties": {
                "description": (
                    f"Debusine Artifact #{artifact.id}: {artifact.category} "
                    f"for {vendor} {codename} {architecture}"
                ),
                "os": vendor,
                "release": codename,
                "debusine_artifact_id": artifact.id,
                "debian_architecture": architecture,
            },
            "templates": {
                "/etc/hostname": {
                    "when": ["create", "copy"],
                    "create_only": False,
                    "template": "hostname.tpl",
                    "properties": {},
                },
                "/etc/hosts": {
                    "when": ["create", "copy"],
                    "create_only": False,
                    "template": "hosts.tpl",
                    "properties": {},
                },
            },
        }

    def build_incus_image(
        self, src_path: Path, workdir: Path, artifact: ArtifactResponse
    ) -> Path:
        """
        Convert an incus environment image into an incus image.

        :param src_path: The incus image to convert.
        :param workdir: A temporary directory to work in and store output in.
        :param artifact: The incus image artifact.

        All the real work is done in prepare_incus_image(); this just opens the
        files.
        """
        log.info("Converting %s %i for Incus", artifact.category, artifact.id)
        dest_path = workdir / "incus.tar.xz"
        with tarfile.open(dest_path, "w:xz") as dest:
            self.prepare_incus_image(src_path, dest, artifact)
        return dest_path

    @abstractmethod
    def prepare_incus_image(
        self,
        source: Path,
        dest: TarFile,
        artifact: ArtifactResponse,
    ) -> None:
        """
        Import the image in source into incus image tarball dest.

        Expected to be overridden to include the image itself.
        """

    def _add_incus_metadata(
        self, dest: TarFile, artifact: ArtifactResponse
    ) -> None:
        """
        Write incus metadata into dest.

        This method is available for use in prepare_incus_image().
        """
        self.add_tar_file(
            dest, "metadata.yaml", yaml.dump(self._image_metadata(artifact))
        )

        self.add_tar_directory(dest, "templates")
        for name, content in TEMPLATES.items():
            self.add_tar_file(dest, f"templates/{name}", content)

    def add_tar_directory(
        self, tarball: TarFile, name: str, mode: int = 0o755
    ) -> None:
        """Add a directory named name to tarball."""
        tarinfo = TarInfo(name=name)
        tarinfo.type = tarfile.DIRTYPE
        tarinfo.mode = mode
        tarball.addfile(tarinfo)

    def add_tar_file(
        self, tarball: TarFile, name: str, data: str, mode: int = 0o644
    ) -> None:
        """Add a file named name to tarball, containing data."""
        encoded = data.encode("utf-8")
        tarinfo = TarInfo(name=name)
        tarinfo.mode = mode
        tarinfo.size = len(encoded)
        tarball.addfile(tarinfo, BytesIO(encoded))

    def add_tar_symlink(
        self,
        tarball: TarFile,
        name: str,
        dest: str,
    ) -> None:
        """Add a symlink named name to tarball."""
        tarinfo = TarInfo(name=name)
        tarinfo.type = tarfile.SYMTYPE
        tarinfo.linkname = dest
        tarball.addfile(tarinfo)


class IncusLXCImageCache(IncusImageCache):
    """Process LXC container images on import."""

    backend = BackendType.INCUS_LXC

    @override
    def prepare_incus_image(
        self, source: Path, dest: TarFile, artifact: ArtifactResponse
    ) -> None:
        """Import the image in source into incus image tarball dest."""
        self._add_incus_metadata(dest, artifact)

        rootfs = Path("rootfs")
        parent_directories = {
            rootfs / "etc": False,
            rootfs / "etc/systemd": False,
            rootfs / "etc/systemd/system": False,
            rootfs / "etc/systemd/system-generators": False,
            rootfs / "etc/systemd/system/multi-user.target.wants": False,
        }
        has_init = False
        init_paths = {
            rootfs / "sbin/init",
            rootfs / "usr/sbin/init",
        }

        # Copy tarball contents into rootfs/
        # We assume it starts with a root directory
        with tarfile.open(source) as source_tar:
            for entry in source_tar:
                dest_path = rootfs / entry.name.lstrip("/")

                if dest_path in init_paths:
                    has_init = True

                if dest_path in parent_directories:
                    parent_directories[dest_path] = True

                dest_entry = entry.replace(name=str(dest_path))
                if entry.isreg():
                    dest.addfile(dest_entry, source_tar.extractfile(entry))
                elif entry.islnk():
                    # Hard-links in tarfiles are stored as an absolute path
                    # (relative to the tar) reference to the shared file
                    dest_entry.linkname = str(
                        rootfs / entry.linkname.lstrip("/")
                    )
                    dest.addfile(dest_entry)
                else:
                    dest.addfile(dest_entry)

        if not has_init:
            raise ImageImportError(
                "Image doesn't contain an init, not bootable in LXC"
            )

        # Create parent directories for the additions below
        for path, present in parent_directories.items():
            if not present:
                self.add_tar_directory(dest, str(path))

        # Enable systemd-networkd, to bring up host0
        self.add_tar_symlink(
            dest,
            (
                "rootfs/etc/systemd/system/multi-user.target.wants/"
                "systemd-networkd.service"
            ),
            "/lib/systemd/system/systemd-networkd.service",
        )
        self.add_tar_symlink(
            dest,
            (
                "rootfs/etc/systemd/system/network-online.target.wants/"
                "systemd-networkd-wait-online.service"
            ),
            "/lib/systemd/system/systemd-networkd-wait-online.service",
        )
        # Install lxc.generator. Contains hacks for systemd < 257
        self.add_tar_file(
            dest,
            "rootfs/etc/systemd/system-generators/lxc",
            importlib.resources.files(__package__)
            .joinpath("data")
            .joinpath("lxc.generator")
            .read_text(encoding="utf-8"),
            mode=0o755,
        )


class IncusVMImageCache(IncusImageCache):
    """Process VM container images on import."""

    backend = BackendType.INCUS_VM

    def convert_raw_image(self, source: Path) -> Path:
        """Convert a raw image to qcow2."""
        workdir = source.parent
        with tarfile.open(source) as tar:
            contents = tar.getnames()
            if len(contents) != 1:
                raise AssertionError(
                    "Raw image tarballs are expected to contain 1 file"
                )
            name = contents[0]
            if "/" in name:
                raise AssertionError(
                    "Raw image tarballs must not include any directories"
                )
            # TODO: Remove filter="data" once we require Python >= 3.14,
            # since it's the new default.
            tar.extract(name, path=workdir, set_attrs=False, filter="data")
            raw = workdir / name

        dest = workdir / "image.qcow2"
        check_call(
            [
                "qemu-img",
                "convert",
                "-f",
                "raw",
                "-O",
                "qcow2",
                str(raw),
                str(dest),
            ]
        )
        raw.unlink()
        return dest

    @override
    def prepare_incus_image(
        self, source: Path, dest: TarFile, artifact: ArtifactResponse
    ) -> None:
        """Import the image in source into incus image tarball dest."""
        self._add_incus_metadata(dest, artifact)

        if artifact.data["image_format"] == "raw":
            source = self.convert_raw_image(source)

        entry = tarfile.TarInfo(name="rootfs.img")
        entry.size = source.stat().st_size
        with source.open("rb") as image:
            dest.addfile(entry, image)

        if artifact.data["image_format"] == "raw":
            source.unlink()


class IncusExecutorMixin:
    """Support the common behaviour between Incus containers and VMs."""

    system_image: ArtifactResponse
    working_directory: Path
    image_cache_class: type[ImageCache]
    image_category: ClassVar[ExecutorImageCategory]
    incus_driver: str
    _image_id: str | None = None

    def __init__(
        self,
        debusine_api: Debusine,
        system_image_id: int,
        workspace: str,
        working_directory: Path,
    ) -> None:
        """
        Instantiate an IncusExecutor.

        :param debusine_api: The object to use the debusine client API.
        :param system_image_id: An artifact ID pointing to the system tarball.
        """
        self._extracted: Path | None = None
        self._image_cache = self.image_cache_class(
            debusine_api, self.image_category, workspace=workspace
        )
        self.system_image = self._image_cache.image_artifact(system_image_id)
        self.working_directory = working_directory

    @classmethod
    def available(cls) -> bool:
        """Determine whether this executor is available for operation."""
        if not utils.is_command_available("incus"):
            return False
        if not utils.is_command_available("autopkgtest-virt-incus"):
            return False

        try:
            output = run_incus_cmd("info")
        except CalledProcessError:
            return False

        info = yaml.safe_load(output)
        drivers = [
            driver.strip()
            for driver in info["environment"]["driver"].split("|")
        ]
        return cls.incus_driver in drivers

    def download_image(self) -> str:
        """
        Make the image available locally.

        Fetch the image from artifact storage, if it isn't already available,
        and make it available locally.

        Return a path to the image or name, as appropriate for the backend.
        """
        self._image_cache.download_image(self.system_image)
        return self.image_name()

    def image_name(self) -> str:
        """Return the alias of the imported image."""
        return f"artifact/{self.system_image.id}"

    def create(self) -> "IncusInstance":
        """
        Create an IncusInstance using the imported image.

        Returns a new, stopped UnshareInstance.
        """
        return IncusInstance(self.image_name())

    def autopkgtest_virt_server(self) -> str:
        """Return the name of the autopkgtest-virt-server for this backend."""
        return "incus"

    @classmethod
    def clean_up_image(cls, artifact_id: int) -> None:
        """Remove incus images from the cache."""
        run_incus_cmd("image", "rm", f"artifact/{artifact_id}")

    @classmethod
    def get_statistics(
        cls, instance: InstanceInterface | None
    ) -> ExecutorStatistics:
        """Return statistics for this executor."""
        statistics = ExecutorStatistics()

        if instance is not None and instance.is_started():
            assert isinstance(instance, IncusInstance)
            assert instance._instance is not None

            free_disk_space: int | None = None
            free_memory: int | None = None
            for family in text_string_to_metric_families(
                subprocess.run(
                    ["incus", "query", "/1.0/metrics"],
                    capture_output=True,
                    check=True,
                    text=True,
                ).stdout
            ):

                def sum_relevant_samples() -> float:
                    return sum(
                        sample.value
                        for sample in family.samples
                        if sample.labels.get("name") == instance._instance
                    )

                match family.name:
                    case "incus_cpu_effective_total":
                        statistics.cpu_count = int(sum_relevant_samples())
                    case "incus_filesystem_avail_bytes":
                        free_disk_space = int(sum_relevant_samples())
                    case "incus_filesystem_size_bytes":
                        statistics.available_disk_space = int(
                            sum_relevant_samples()
                        )
                    case "incus_memory_MemFree_bytes":
                        free_memory = int(sum_relevant_samples())
                    case "incus_memory_MemTotal_bytes":
                        statistics.available_memory = int(
                            sum_relevant_samples()
                        )

            if (
                statistics.available_disk_space is not None
                and free_disk_space is not None
            ):
                statistics.disk_space = (
                    statistics.available_disk_space - free_disk_space
                )
            if (
                statistics.available_memory is not None
                and free_memory is not None
            ):
                # TODO: This is an approximation; in particular, it doesn't take
                # swap into account.  Incus's metrics don't seem to include
                # available swap space, so we'd probably need to do something
                # more complicated to sum all the relevant categories of used
                # memory.
                statistics.memory = statistics.available_memory - free_memory

        return statistics


class IncusLXCExecutor(
    IncusExecutorMixin, ExecutorInterface, backend_name=BackendType.INCUS_LXC
):
    """Support the Incus LXC Container executor."""

    image_cache_class = IncusLXCImageCache
    image_category = ExecutorImageCategory.TARBALL
    incus_driver = "lxc"

    @override
    def autopkgtest_virt_args(self) -> list[str]:
        image_id = self.image_name()
        return [image_id, "--", "--profile", "debusine"]


class IncusVMExecutor(
    IncusExecutorMixin, ExecutorInterface, backend_name=BackendType.INCUS_VM
):
    """Support the Incus VM executor."""

    image_cache_class = IncusVMImageCache
    image_category = ExecutorImageCategory.IMAGE
    incus_driver = "qemu"

    @override
    def autopkgtest_virt_args(self) -> list[str]:
        """Generate the arguments to drive an autopkgtest-virt-server."""
        image_id = self.image_name()
        return ["--vm", image_id, "--", "--profile", "debusine"]


class IncusInstance(InstanceInterface):
    """Support instances of the Incus executor."""

    _image: str
    _instance: str | None = None

    def __init__(self, image: str) -> None:
        """Initialize the object."""
        super().__init__()
        self._image = image

    @override
    def is_started(self) -> bool:
        return self._instance is not None

    @classmethod
    def _generate_instance_name(cls) -> str:
        """Generate an unused incus instance name."""
        while True:
            rnd = [random.choice(string.ascii_lowercase) for i in range(6)]
            candidate = "debusine-" + "".join(rnd)
            try:
                run_incus_cmd("info", candidate)
            except CalledProcessError:
                return candidate

    @override
    def do_start(self) -> None:
        """Start an ephemeral instance."""
        name = self._generate_instance_name()
        run_incus_cmd(
            "launch", "--ephemeral", "--profile", "debusine", self._image, name
        )
        self._instance = name
        self._spin_until_started()

    def _spin_until_started(
        self, sleep_time: float = 0.2, timeout: float = 60.0
    ) -> None:
        """Wait for an instance to become ready."""
        assert self._instance is not None
        start_time = tm.monotonic()
        end_time = start_time + timeout
        while tm.monotonic() < end_time:
            p = subprocess.run(
                [
                    "incus",
                    "exec",
                    self._instance,
                    "systemctl",
                    "is-system-running",
                ],
                capture_output=True,
                text=True,
            )
            stdout = p.stdout.strip()
            stderr = p.stderr.strip()
            if stdout in ("running", "degraded", "maintenance"):
                return
            elif stdout in ("initializing", "starting"):
                pass
            elif (
                p.returncode == 1
                and stderr == "Error: VM agent isn't currently running"
            ):
                pass
            elif p.returncode == 1 and stderr.startswith(
                "Failed to connect to system scope bus via local transport:"
            ):
                pass
            else:
                log.warning(
                    "Unexpected response from exec systemctl "
                    "is-system-running: returncode: %i, stdout: %s, stderr: %s",
                    p.returncode,
                    p.stdout,
                    p.stderr,
                )
            tm.sleep(sleep_time)
        raise InstanceNotRunning(
            f"Incus Instance {self._instance} hasn't booted "
            "or is missing incus-agent"
        )

    @override
    def do_stop(self) -> None:
        """Stop the instance."""
        assert self._instance is not None
        run_incus_cmd("stop", self._instance)
        self._instance = None

    @override
    def do_restart(self) -> None:
        """Restart the instance."""
        assert self._instance is not None
        run_incus_cmd("restart", self._instance)
        self._spin_until_started()

    @override
    def do_file_push(
        self, source: Path, target: PurePath, uid: int, gid: int, mode: int
    ) -> None:
        """
        Copy a file into the environment.

        source is a regular file.
        target is the target file-name within an existing directory in the
        instance.

        Files are owned by the same owner as their parent directory.
        Timestamps are not expected to be retained.
        """
        assert self._instance is not None
        assert target.is_absolute()
        run_incus_cmd(
            "file",
            "push",
            "--uid",
            str(uid),
            "--gid",
            str(gid),
            "--mode",
            f"{mode:o}",
            str(source),
            self._instance + str(target),
        )

    @override
    def do_file_pull(self, source: PurePath, target: Path) -> None:
        """
        Copy a file out of the environment.

        source is a regular file.
        target is the target file-name within an existing directory on the
        host.

        Timestamps are not expected to be retained.
        """
        assert self._instance is not None
        assert source.is_absolute()
        run_incus_cmd("file", "pull", self._instance + str(source), str(target))

    @override
    def do_directory_push(
        self, source: Path, target: PurePath, uid: int, gid: int
    ) -> None:
        """
        Copy a directory (recursively) into the environment.

        source is a directory.
        target is an existing directory that source will be copied into.

        Only regular files are supported.
        Timestamps are not expected to be retained.
        New directories and files are owned by the same owner as their parent
        directory.
        """
        assert self._instance is not None
        assert target.is_absolute()
        run_incus_cmd(
            "file",
            "push",
            "--recursive",
            str(source),
            self._instance + str(target),
        )
        run_incus_cmd(
            "exec",
            self._instance,
            "--",
            "chown",
            "-R",
            f"{uid}:{gid}",
            str(target / source.name),
        )

    @overload
    def do_run(
        self,
        args: list[str],
        text: Literal[True],
        cwd: str | os.PathLike[str] | None = None,
        run_as_root: bool = False,
        **kwargs: Any,
    ) -> CompletedProcess[str]: ...

    @overload
    def do_run(
        self,
        args: list[str],
        text: Literal[False] | None = None,
        cwd: str | os.PathLike[str] | None = None,
        run_as_root: bool = False,
        **kwargs: Any,
    ) -> CompletedProcess[bytes]: ...

    @override
    def do_run(
        self,
        args: list[str],
        text: Literal[False] | Literal[True] | None = None,
        cwd: str | os.PathLike[str] | None = None,
        run_as_root: bool = False,
        **kwargs: Any,
    ) -> CompletedProcess[AnyStr]:
        assert self._instance is not None
        cmd = ["incus", "exec", self._instance]

        if cwd is not None:
            # --cwd defaults to /root, so interpret relative paths as
            # relative to that.
            if not PurePath(cwd).is_absolute():
                cwd = PurePath("/root", cwd)
            cmd += ["--cwd", str(cwd)]

        if not run_as_root:
            uid = self.create_user()
            cmd += ["--user", str(uid)]

        cmd.append("--")
        cmd += args

        return subprocess.run(cmd, text=text, **kwargs)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for Redis Prometheus Values."""

import datetime as dt
import time as tm
from typing import cast

from django.test import SimpleTestCase
from prometheus_client.samples import Exemplar

from debusine.server.prometheus_client.redis import (
    _daemon_threads,
    _live_metrics,
    private_fake_redis,
    redis_client,
)
from debusine.server.prometheus_client.values import (
    MULTIPROCESS_MODE_T,
    RedisValue,
    Value,
)


class RedisValueProtocol(Value):
    """Extend the Value protocol to give us access to RedisValue internals."""

    _typ: str
    _metric_name: str
    _name: str
    _labelnames: list[str]
    _labelvalues: list[str]
    _help_text: str
    _multiprocess_mode: MULTIPROCESS_MODE_T
    _expiry: dt.timedelta | None
    _key: str


@private_fake_redis()
class ValueTestCase(SimpleTestCase):
    """Test Redis values."""

    pid: int = 123

    def create_value(
        self,
        metric_name: str = "test",
        name: str | None = None,
        type_: str = "counter",
        labelnames: list[str] | None = None,
        labelvalues: list[str] | None = None,
        multiprocess_mode: MULTIPROCESS_MODE_T = "",
    ) -> RedisValueProtocol:
        value = RedisValue(lambda: self.pid)(
            type_,
            metric_name,
            name or metric_name + "_total",
            labelnames or [],
            labelvalues or [],
            "Help Text",
            multiprocess_mode=multiprocess_mode,
        )
        return cast(RedisValueProtocol, value)

    def test_initializes_value(self) -> None:
        value = self.create_value()
        self.assertEqual(value.get(), 0.0)

    def test_sets_and_gets_value(self) -> None:
        value = self.create_value()
        value.set(5)
        self.assertEqual(value.get(), 5.0)

    def test_inc_value(self) -> None:
        value = self.create_value()
        value.inc(3)
        value.inc(5)
        self.assertEqual(value.get(), 8.0)

    def test_get_missing_value(self) -> None:
        value = self.create_value()
        redis_client().delete(value._key)
        self.assertEqual(value.get(), 0.0)

    def test_exemplars_not_implemented(self) -> None:
        value = self.create_value("test4")
        with self.assertRaises(NotImplementedError):
            value.set_exemplar(Exemplar(labels={}, value=0.0, timestamp=0.0))
        self.assertIsNone(value.get_exemplar())

    def test_differentiated_by_name(self) -> None:
        v1 = self.create_value("value1")
        v2 = self.create_value("value2")
        v1.set(1)
        v2.set(2)
        self.assertEqual(v1.get(), 1.0)
        self.assertEqual(v2.get(), 2.0)

    def test_differentiated_by_labels(self) -> None:
        v1 = self.create_value("value3", labelnames=["a"], labelvalues=["1"])
        v2 = self.create_value("value3", labelnames=["a"], labelvalues=["2"])
        v1.set(1)
        v2.set(2)
        self.assertEqual(v1.get(), 1.0)
        self.assertEqual(v2.get(), 2.0)

    def test_multiprocess_mode_mostrecent(self) -> None:
        with self.assertRaises(NotImplementedError):
            self.create_value(type_="gauge", multiprocess_mode="mostrecent")

    def test_multiprocess_mode_counter(self) -> None:
        with self.assertRaises(AssertionError):
            self.create_value(type_="counter", multiprocess_mode="liveall")

    def test_multiprocess_mode(self) -> None:
        value = self.create_value(type_="gauge", multiprocess_mode="all")
        self.assertEqual(value._labelnames, ["pid"])
        self.assertEqual(value._labelvalues[-1], "123")
        self.assertIsNone(value._expiry)
        self.assertLess(redis_client().expiretime(value._key), 0)

    def test_multiprocess_mode_live(self) -> None:
        value = self.create_value(type_="gauge", multiprocess_mode="liveall")
        unixtime = tm.time()
        self.assertEqual(value._labelnames, ["pid"])
        self.assertEqual(value._labelvalues[-1], "123")
        self.assertEqual(value._expiry, dt.timedelta(seconds=20))
        expiretime = redis_client().expiretime(value._key)
        self.assertGreater(expiretime, unixtime)
        self.assertLessEqual(expiretime, unixtime + 20)

        self.assertIn("123", _daemon_threads)
        self.assertIn("123", _live_metrics)
        self.assertIn(value._key, _live_metrics["123"])

    def test_live_inc_updates_expiry(self) -> None:
        value = self.create_value(type_="gauge", multiprocess_mode="liveall")
        unixtime = tm.time()
        redis_client().persist(value._key)
        self.assertLess(redis_client().expiretime(value._key), 0)

        value.inc(1)
        self.assertGreater(redis_client().expiretime(value._key), unixtime)

    def test_live_set_updates_expiry(self) -> None:
        value = self.create_value(type_="gauge", multiprocess_mode="liveall")
        unixtime = tm.time()
        redis_client().persist(value._key)
        self.assertLess(redis_client().expiretime(value._key), 0)

        value.set(1)
        self.assertGreater(redis_client().expiretime(value._key), unixtime)

    def test_live_shares_daemon_thread(self) -> None:
        self.create_value("value1", type_="gauge", multiprocess_mode="liveall")
        self.create_value("value2", type_="gauge", multiprocess_mode="liveall")

    def test_multiprocess_pid_change(self) -> None:
        self.pid = 1

        value = self.create_value(type_="gauge", multiprocess_mode="all")
        self.assertEqual(value._labelnames[-1], "pid")
        self.assertEqual(value._labelvalues[-1], "1")
        value.inc(1)
        self.assertEqual(value.get(), 1.0)

        self.pid = 2
        value.inc(1)
        self.assertEqual(value._labelvalues[-1], "2")
        self.assertEqual(value.get(), 1.0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Helper mixin for the integration tests."""

import os
import socket
import subprocess
import tempfile
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from textwrap import dedent
from typing import Any

import requests
import yaml
from debian.deb822 import Deb822

from debusine.artifacts.local_artifact import BinaryPackage, SourcePackage
from debusine.artifacts.models import ArtifactCategory
from debusine.client.config import ConfigHandler
from debusine.tasks.models import BackendType
from debusine.test import TestCase
from utils.client import Client
from utils.server import DebusineServer


class IntegrationTestHelpersMixin(TestCase):
    """Utility methods for the integration tests."""

    @staticmethod
    @contextmanager
    def _temporary_directory() -> Generator[Path]:
        with tempfile.TemporaryDirectory(
            prefix="debusine-integration-tests-"
        ) as temp_directory:
            yield Path(temp_directory)

    @staticmethod
    def get_client_token_key() -> str:
        """Get our API token key."""
        config = ConfigHandler(
            server_name="integration-test"
        ).server_configuration()
        assert config.api_token is not None
        return config.api_token

    @classmethod
    def make_apt_environment(self, apt_path: Path) -> dict[str, str]:
        """Make a suitable process environment for running APT commands."""
        return {**os.environ, "APT_CONFIG": str(apt_path / "etc/apt/apt.conf")}

    @classmethod
    @contextmanager
    def apt_indexes(
        cls,
        suite_name: str,
        url: str = "https://deb.debian.org/debian",
        signed_by: (
            str | None
        ) = "/usr/share/keyrings/debian-archive-keyring.gpg",
        architectures: list[str] | None = None,
        trusted: bool = False,
        auth: tuple[str, str] | None = None,
    ) -> Generator[Path]:
        """Fetch APT indexes for a suite."""
        if architectures is None:
            architectures = [
                subprocess.check_output(
                    ["dpkg", "--print-architecture"], text=True
                ).strip()
            ]

        with cls._temporary_directory() as apt_path:
            (apt_path / "etc/apt/apt.conf.d").mkdir(parents=True)
            (apt_path / "etc/apt/preferences.d").mkdir(parents=True)
            (apt_path / "etc/apt/sources.list.d").mkdir(parents=True)
            (apt_path / "var/lib/apt/lists/partial").mkdir(parents=True)
            (apt_path / "etc/apt/apt.conf").write_text(
                dedent(
                    f"""\
                    APT::Architecture "{architectures[0]}";
                    APT::Architectures "{','.join(architectures)}";
                    Dir "{apt_path}";
                    """
                )
            )
            source = {
                "Types": "deb deb-src",
                "URIs": url,
                "Suites": suite_name,
                "Components": "main",
            }
            if signed_by is not None:
                if "\n" in signed_by:
                    # A multi-line Signed-By is an embedded public key.
                    signed_by = "\n" + "\n".join(
                        f" {line}" if line else " ."
                        for line in signed_by.splitlines()
                    )
                source["Signed-By"] = signed_by
            if trusted:
                source["Trusted"] = "yes"
            (apt_path / "etc/apt/sources.list.d/test.sources").write_text(
                str(Deb822(source))
            )
            if auth is not None:
                auth_file = apt_path / "etc/apt/auth.conf.d/test.conf"
                auth_file.parent.mkdir(parents=True)
                auth_file.write_text(
                    dedent(
                        f"""\
                        machine {url}
                        login {auth[0]}
                        password {auth[1]}
                        """
                    )
                )
                auth_file.chmod(0o600)
            subprocess.check_call(
                ["apt-get", "update"], env=cls.make_apt_environment(apt_path)
            )
            yield apt_path

    @classmethod
    def _create_artifact(
        cls,
        artifact_type: ArtifactCategory,
        data: dict[str, Any],
        files: list[Path],
        base_directory: Path,
    ) -> int:
        """
        Create an artifact.

        :param artifact_type: type of artifact (e.g. debian:source-package,
          debian:binary-package).
        :param data: type-specific data to attach to the artifact.
        :param files: list of paths to files to attach to the artifact.
        :param base_directory: base directory for files to upload (to allow
          calculating relative paths)
        """
        with cls._temporary_directory() as temp_path:
            data_yaml = temp_path / "artifact-data.yaml"
            data_yaml.write_text(yaml.safe_dump(data))

            artifact_id = Client.execute_command(
                "artifact",
                "create",
                artifact_type,
                "--workspace",
                "System",
                "--data",
                data_yaml,
                "--upload-base-directory",
                base_directory,
                "--upload",
                *files,
            )["artifact_id"]
            assert isinstance(artifact_id, int)
            return artifact_id

    @classmethod
    def _create_apt_artifact(
        cls,
        apt_path: Path,
        package_name: str,
        apt_get_command: str,
        artifact_type: ArtifactCategory,
    ) -> int:
        """
        Download files using apt and create an artifact from them.

        :param suite_name: suite to download from.
        :param package_name: package contained in the artifact.
        :param apt_get_command: e.g. "source", "download".
        :param artifact_type: type of artifact (e.g. debian:source-package,
          debian:binary-package).
        """
        with cls._temporary_directory() as temp_path:
            subprocess.check_call(
                ["apt-get", apt_get_command, "--download-only", package_name],
                cwd=temp_path,
                env=cls.make_apt_environment(apt_path),
            )

            files = list(temp_path.iterdir())

            if artifact_type == ArtifactCategory.SOURCE_PACKAGE:
                [dsc_file] = [
                    file for file in files if file.name.endswith(".dsc")
                ]
                version = dsc_file.name.removesuffix(".dsc").split("_")[1]
                src_artifact = SourcePackage.create(
                    name=package_name, version=version, files=files
                )
                data = src_artifact.data.model_dump()
            elif artifact_type == ArtifactCategory.BINARY_PACKAGE:
                bin_artifact = BinaryPackage.create(file=files[0])
                data = bin_artifact.data.model_dump()
            else:
                raise NotImplementedError(
                    f"cannot generate artifact data for {artifact_type}"
                )

            return cls._create_artifact(artifact_type, data, files, temp_path)

    @classmethod
    def create_artifact_source(cls, apt_path: Path, package_name: str) -> int:
        """Create a source artifact with hello files."""
        return cls._create_apt_artifact(
            apt_path, package_name, "source", ArtifactCategory.SOURCE_PACKAGE
        )

    @classmethod
    def create_artifact_binary(cls, apt_path: Path, package_name: str) -> int:
        """Create a binary artifact."""
        return cls._create_apt_artifact(
            apt_path, package_name, "download", ArtifactCategory.BINARY_PACKAGE
        )

    @classmethod
    def create_artifact_upload(
        cls, apt_path: Path, package_names: list[str]
    ) -> int:
        """Create an upload artifact containing some binary packages."""
        with cls._temporary_directory() as temp_path:
            for package_name in package_names:
                subprocess.check_call(
                    ["apt-get", "download", "--download-only", package_name],
                    cwd=temp_path,
                    env=cls.make_apt_environment(apt_path),
                )

            files = sorted(temp_path.iterdir())
            changes_file = temp_path / f"{package_names[0]}.changes"
            cls.write_changes_file(changes_file, files)

            artifact_id = Client.execute_command(
                "artifact",
                "import-debian",
                "--workspace",
                "System",
                changes_file,
            )["artifact_id"]
            assert isinstance(artifact_id, int)
            return artifact_id

    @classmethod
    def create_artifact_build_logs(cls, filename: str, contents: str) -> int:
        """
        Create a build-log artifact.

        :param filename: filename that is being created
        :param contents: contents of the file being created
        """
        with cls._temporary_directory() as temp_path:
            artifact_type = ArtifactCategory.PACKAGE_BUILD_LOG

            (temp_path / filename).write_text(contents)
            files = list(temp_path.iterdir())

            data = {
                "source": "test",
                "version": "1.0",
                "filename": filename,
            }

            return cls._create_artifact(artifact_type, data, files, temp_path)

    @classmethod
    def create_artifact_signing_input(
        cls, binary_package_name: str, files: dict[str, bytes]
    ) -> int:
        """
        Create a signing-input artifact.

        :param binary_package_name: name of the binary package this represents.
        :param files: mapping of filenames to contents, each of which is to
          be signed
        :return: the created artifact ID
        """
        with cls._temporary_directory() as temp_path:
            artifact_type = ArtifactCategory.SIGNING_INPUT

            for filename, contents in files.items():
                (temp_path / filename).parent.mkdir(parents=True, exist_ok=True)
                (temp_path / filename).write_bytes(contents)

            return cls._create_artifact(
                artifact_type,
                {
                    "binary_package_name": binary_package_name,
                },
                [temp_path / filename for filename in files],
                temp_path,
            )

    @staticmethod
    def print_work_request_debug_logs(work_request_id: int) -> None:
        """Print work request's log's - used for debugging."""
        work_request = Client.execute_command(
            "work-request", "show", work_request_id
        )

        for artifact in work_request["artifacts"]:
            if artifact["category"] == ArtifactCategory.WORK_REQUEST_DEBUG_LOGS:
                for path, file_information in artifact["files"].items():
                    print("------ FILE:", path)
                    print(">>>>>>>>>>>")
                    print(requests.get(file_information["url"]).text)
                    print("<<<<<<<<<<<")

    @staticmethod
    def create_system_images_mmdebstrap(
        suites: list[str], backend: str = BackendType.UNSHARE
    ) -> None:
        """
        Create debian:system-image artifacts and add them to a collection.

        :param suites: suites of the system images
        """
        template_data = {
            "static_parameters": {
                "vendor": "debian",
                "targets": [
                    {
                        "codenames": suites,
                        "backends": [backend],
                        "architectures": [
                            subprocess.check_output(
                                ["dpkg", "--print-architecture"], text=True
                            ).strip()
                        ],
                        "mmdebstrap_template": {
                            "bootstrap_options": {"variant": "apt"},
                            "bootstrap_repositories": [
                                {
                                    "mirror": "http://deb.debian.org/debian",
                                    "components": ["main"],
                                }
                            ],
                        },
                    }
                ],
            }
        }

        workflow_template_result = Client.execute_command(
            "workflow-template",
            "create",
            "update_environments",
            f"mmdebstrap-environment-{'-'.join(suites)}",
            stdin=yaml.safe_dump(template_data),
        )
        assert workflow_template_result.returncode == 0

        workflow_id = Client.execute_command(
            "workflow",
            "start",
            "--yaml",
            f"mmdebstrap-environment-{'-'.join(suites)}",
            stdin=yaml.safe_dump({}),
        )["id"]

        # The worker should get the new work request and start executing it
        assert Client.wait_for_work_request_completed(
            workflow_id, "success", timeout=1800 * len(suites)
        )

    def create_suite(
        self,
        suite_name: str,
        *,
        workspace_name: str = "System",
        architectures: list[str],
        base_workflow_template: str | None = None,
    ) -> None:
        """Create a suite collection."""
        command = ["archive", "suite", "create", "--workspace", workspace_name]
        for architecture in architectures:
            command += ["--architecture", architecture]
        if base_workflow_template is not None:
            command += ["--base-workflow-template", base_workflow_template]
        command.append(suite_name)
        result = Client.execute_command(*command)
        self.assertEqual(result.returncode, 0)

    def add_suite_to_archive(
        self, suite_name: str, *, workspace_name: str = "System"
    ) -> None:
        """Add a suite to our workspace's archive."""
        # TODO: We don't have a sensible command to do this yet.
        result = DebusineServer.execute_command(
            "shell",
            "-v",
            "0",
            "-c",
            dedent(
                f"""\
                from django.conf import settings

                from debusine.db.permissioncontext import PermissionContext
                from debusine.artifacts.models import *
                from debusine.db.models import *

                user = User.objects.get(username="test-user")
                pc = PermissionContext.for_ui_user(user)
                workspace = Workspace.objects.get(
                    scope__name=settings.DEBUSINE_DEFAULT_SCOPE,
                    name={workspace_name!r},
                )
                archive = workspace.get_singleton_collection(
                    pc=pc, category=CollectionCategory.ARCHIVE
                )
                suite = workspace.get_collection(
                    pc=pc,
                    category=CollectionCategory.SUITE,
                    name={suite_name!r},
                )
                archive.manager.add_collection(suite, user=user)
                """
            ),
        )
        self.assertEqual(result.returncode, 0)

    def get_archive_signing_key(
        self,
        workspace_name: str = "System",
        suite_name: str | None = None,
        auth: tuple[str, str] | None = None,
    ) -> str:
        """Get the signing key for our workspace's archive, or a suite."""
        url = f"https://deb.{socket.getfqdn()}/debusine/{workspace_name}"
        if suite_name is not None:
            url += f"/dists/{suite_name}"
        url += "/signing-keys.asc"
        response = requests.get(url, auth=auth)
        response.raise_for_status()
        return response.text



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data models for db users and authentication."""

import abc
import datetime as dt
import enum
import hashlib
import re
import secrets
from enum import StrEnum
from typing import (
    Annotated,
    Any,
    Literal,
    Optional,
    Self,
    TYPE_CHECKING,
    TypeAlias,
    cast,
    override,
)

import pgtrigger
import pydantic
from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.contrib.auth.models import UserManager as DjangoUserManager
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.core.validators import MaxLengthValidator, MinLengthValidator
from django.db import models
from django.db.models import (
    CheckConstraint,
    Exists,
    F,
    OuterRef,
    Q,
    QuerySet,
    UniqueConstraint,
    Value,
)
from django.db.models.functions import Coalesce
from django.urls import reverse
from django.utils import timezone

from debusine.db.models import permissions
from debusine.db.models.permissions import (
    Allow,
    Role,
    get_resource_scope,
    permission_check,
    permission_filter,
    resolve_role,
)
from debusine.db.models.scopes import Scope
from debusine.db.models.workspaces import Workspace
from debusine.db.permissioncontext import PermissionContext
from debusine.server import notifications
from debusine.utils.typing_utils import copy_signature_from

if TYPE_CHECKING:
    from django.http import HttpRequest
    from django_stubs_ext.db.models import TypedModelMeta

    from debusine.db.models.work_requests import WorkRequest
    from debusine.web.views.ui.auth import GroupMembershipUI, GroupUI, UserUI

    # https://github.com/typeddjango/django-stubs/issues/2112
    AbstractUserMeta = TypedModelMeta
else:
    TypedModelMeta = object
    AbstractUserMeta = AbstractUser.Meta


class TokenQuerySet[A](QuerySet["Token", A]):
    """Custom QuerySet for Token."""

    def get_tokens(
        self,
        username: str | None = None,
        key: str | None = None,
        include_expired: bool = False,
    ) -> Self:
        """
        Return all the tokens filtered by a specific owner and/or token.

        To avoid filtering by owner or token set them to None
        """
        tokens = self

        if username:
            tokens = tokens.filter(user__username=username)

        if key:
            token_hash = hashlib.sha256(key.encode()).hexdigest()
            tokens = tokens.filter(hash=token_hash)

        if not include_expired:
            tokens = tokens.exclude(expire_at__lt=timezone.now())

        return tokens

    def get_token_or_none(self, token_key: str) -> Optional["Token"]:
        """Return the token with token_key or None."""
        assert isinstance(token_key, str)

        token_hash = hashlib.sha256(token_key.encode()).hexdigest()

        try:
            return cast(
                Token | None,
                (
                    self.select_related(
                        "worker",
                        "activating_worker",
                        "work_request__worker",
                        "work_request__created_by",
                        "work_request__workflow_root__workflow_template",
                    )
                    .exclude(expire_at__lt=timezone.now())
                    .get(hash=token_hash)
                ),
            )
        except Token.DoesNotExist:
            return None

    def expired(self, at: dt.datetime) -> Self:
        """
        Return queryset with tokens that have expired.

        :param at: datetime to check if the tokens are expired.
        :return: tokens whose ``expire_at`` is before the given datetime.
        """
        return self.filter(expire_at__lt=at)


class TokenManager(models.Manager["Token"]):
    """Manager for Token model."""

    def create_worker_activation(self) -> "Token":
        """Create a token suitable for use in worker activation."""
        # These are currently hardcoded to expire in 15 minutes, which
        # should be enough for most providers to be able to launch an
        # instance.
        return self.create(
            token_type=Token.Types.WORKER_ACTIVATION,
            expire_at=timezone.now() + dt.timedelta(minutes=15),
            enabled=True,
        )

    def create_work_request_token(self, work_request: "WorkRequest") -> "Token":
        """Create a limited token for use in a single work request."""
        # Avoid circular imports
        from debusine.db.models.work_requests import WorkRequest

        if work_request.status not in (
            WorkRequest.Statuses.PENDING,
            WorkRequest.Statuses.RUNNING,
        ):
            raise AssertionError(
                f"Cannot issue a token for a WorkRequest that has status "
                f"{work_request.status}"
            )
        if work_request.worker is None:
            raise AssertionError(
                "Cannot issue a token for a WorkRequest that is not assigned "
                "to a worker."
            )
        # Rotate any existing token. We only permit one task per worker & work
        # request, simply delete any stale tokens.
        # This will need to change to support concurrency > 1
        Token.objects.filter(token_type=Token.Types.WORK_REQUEST).filter(
            Q(work_request=work_request)
            | Q(work_request__worker=work_request.worker)
        ).delete()

        return self.create(
            expire_at=timezone.now() + dt.timedelta(days=7),
            enabled=True,
            token_type=Token.Types.WORK_REQUEST,
            work_request=work_request,
        )

    def create_external_user_token(
        self,
        *,
        username: str,
        created_by: "User | None" = None,
        comment: str = "",
    ) -> "Token":
        """Create a token for an external user."""
        return self.create(
            token_type=Token.Types.EXTERNAL_USER,
            external_username=username,
            created_by=created_by,
            explicit_grants_only=True,
            comment=comment,
        )

    @override
    def get_queryset(self) -> TokenQuerySet[Any]:
        """Use the custom QuerySet."""
        return TokenQuerySet(self.model, using=self._db)


class _TokenTypes(models.TextChoices):
    #: Token used for worker activation only
    WORKER_ACTIVATION = "worker-activation", "Worker activation"
    #: Token used by the worker to connect to Debusine
    WORKER = "worker", "Worker"
    #: Limited token for use in a work request
    WORK_REQUEST = "work-request", "Work request"
    #: User authentication token for debusine client
    USER = "user", "User"
    EXTERNAL_USER = "external-user", "User without a Debusine identity"


class Token(models.Model):
    """
    Database model of a token.

    A token contains a key and other related data. It's used as a shared
    key between debusine server and clients (workers).

    This token model is very similar to rest_framework.authtoken.models.Token.
    The bigger difference is that debusine's token's owner is a CharField,
    the rest_framework owner is a OneToOne foreign key to a user.

    Database-wise we don't store the token itself, but a hash of the token.
    :py:func:`TokenManager.get_token_or_none` can be used to check a provided
    token key against the database.
    """

    Types: TypeAlias = _TokenTypes

    token_type = models.CharField(choices=Types.choices)

    hash = models.CharField(
        max_length=64,
        unique=True,
        verbose_name='Hexadecimal hash, length is 64 chars',
        validators=[MaxLengthValidator(64), MinLengthValidator(64)],
    )
    _worker_id = models.BigIntegerField(
        null=True,
        editable=False,
        db_column="worker_id",
        help_text="Mirror of Worker.token; only for use in constraints",
    )
    _activating_worker_id = models.BigIntegerField(
        null=True,
        editable=False,
        db_column="activating_worker_id",
        help_text=(
            "Mirror of Worker.activation_token; only for use in constraints"
        ),
    )

    # `user` is used to authenticate as a Debusine user, while
    # `external_username` is used in contexts where we allow authenticating
    # using an externally-provisioned username.
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        blank=True,
        on_delete=models.PROTECT,
    )
    # max_length, help_text, and validators match AbstractUser.username.
    external_username = models.CharField(
        max_length=150,
        blank=True,
        help_text=(
            "Username to be used with an external user token;"
            " letters, digits, and @/./+/-/_ only."
        ),
        validators=[UnicodeUsernameValidator()],
    )
    work_request = models.OneToOneField(
        "WorkRequest",
        null=True,
        blank=True,
        related_name="work_request_token",
        on_delete=models.PROTECT,
    )

    comment = models.CharField(
        max_length=100,
        default='',
        verbose_name='Reason that this token was created',
        blank=True,
    )
    created_by = models.ForeignKey(
        "db.User",
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name="created_tokens",
    )
    created_at = models.DateTimeField(auto_now_add=True)
    expire_at = models.DateTimeField(null=True, blank=True)

    explicit_grants_only = models.BooleanField(null=False, default=False)

    enabled = models.BooleanField(default=False)

    last_seen_at = models.DateTimeField(
        blank=True,
        null=True,
        help_text="Last time that the token was used",
    )

    key: str

    objects = TokenManager.from_queryset(TokenQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"
        constraints = [
            # Only the defined token types are allowed.  (By default, Django
            # only enforces this at the form level.)
            CheckConstraint(
                name="%(app_label)s_%(class)s_valid_type",
                check=Q(token_type__in=sorted(_TokenTypes)),
            ),
            # The token type must be consistent with its associated fields.
            CheckConstraint(
                name="%(app_label)s_%(class)s_consistent_type",
                check=(
                    (
                        Q(token_type=_TokenTypes.WORKER_ACTIVATION)
                        | Q(_activating_worker_id__isnull=True)
                    )
                    & (
                        Q(token_type=_TokenTypes.WORKER)
                        | Q(_worker_id__isnull=True)
                    )
                    & (
                        Q(token_type=_TokenTypes.WORK_REQUEST)
                        | Q(work_request_id__isnull=True)
                    )
                    & (Q(token_type=_TokenTypes.USER) | Q(user__isnull=True))
                    & (
                        Q(token_type=_TokenTypes.EXTERNAL_USER)
                        | Q(external_username="")
                    )
                ),
            ),
            # External user tokens must use the explicit-grants-only mode,
            # since without explicit grants they would be no different from
            # anonymous access.  User tokens may use that mode.  Other
            # tokens may not.
            CheckConstraint(
                name="%(app_label)s_%(class)s_consistent_grants",
                check=(
                    (
                        ~Q(token_type=_TokenTypes.EXTERNAL_USER)
                        | Q(explicit_grants_only=True)
                    )
                    & (
                        Q(
                            token_type__in=(
                                _TokenTypes.USER,
                                _TokenTypes.EXTERNAL_USER,
                            )
                        )
                        | Q(explicit_grants_only=False)
                    )
                ),
            ),
        ]
        triggers = [
            # Each token type has another corresponding non-nullable field,
            # but we have to wait for the end of the transaction before
            # checking this because in some cases there's also a check
            # constraint on the other side of a foreign key and it therefore
            # requires multiple statements for everything to be consistent.
            #
            # We must fetch the current state of the row at the start of
            # this trigger function, since the value of NEW is fixed when
            # the trigger is scheduled.
            pgtrigger.Trigger(
                name="db_token_required_type_parameters",
                when=pgtrigger.After,
                operation=pgtrigger.Insert | pgtrigger.Update,
                timing=pgtrigger.Deferred,
                func=" ".join(
                    """
                    SELECT * FROM db_token INTO NEW WHERE id = db_token.id;
                    IF NEW.token_type = 'worker-activation'
                       AND NEW.activating_worker_id IS NULL THEN
                        RAISE EXCEPTION
                            'Worker activation token % has no activating worker',
                            NEW.id;
                    ELSIF NEW.token_type = 'worker'
                          AND NEW.worker_id IS NULL THEN
                        RAISE EXCEPTION 'Worker token % has no worker', NEW.id;
                    ELSIF NEW.token_type = 'work-request'
                          AND NEW.work_request_id IS NULL THEN
                        RAISE EXCEPTION
                            'Work request token % has no work request',
                            NEW.id;
                    ELSIF NEW.token_type = 'user' AND NEW.user_id IS NULL THEN
                        RAISE EXCEPTION 'User token % has no user', NEW.id;
                    ELSIF NEW.token_type = 'external-user'
                          AND new.external_username = '' THEN
                        RAISE EXCEPTION
                            'External user token % has no username', NEW.id;
                    ELSE
                        RETURN NULL;
                    END IF;
                    """.split()  # noqa: E501
                ),
            )
        ]

    @override
    def __str__(self) -> str:
        """Return the hash of the Token."""
        return self.hash

    @override
    def save(self, *args: Any, **kwargs: Any) -> None:
        """Save the token. If it's a new token it generates a key."""
        if not self.hash:
            self.key = self._generate_key()
            self.hash = self._generate_hash(self.key)
        super().save(*args, **kwargs)

    def enable(self) -> None:
        """Enable the token and save it."""
        self.enabled = True
        self.save()

    def disable(self) -> None:
        """Disable the token and save it."""
        self.enabled = False
        self.save()

        notifications.notify_worker_token_disabled(self)

    @classmethod
    def _generate_key(cls) -> str:
        """Create and return a key."""
        return secrets.token_hex(32)

    @classmethod
    def _generate_hash(cls, secret: str) -> str:
        """Hash the given secret."""
        return hashlib.sha256(secret.encode()).hexdigest()

    def get_username(self) -> str:
        """Get a meaningful username for the token, if possible."""
        match self.token_type:
            case Token.Types.USER:
                assert self.user is not None
                return self.user.username
            case Token.Types.EXTERNAL_USER:
                return self.external_username
            case _:
                return ""


SYSTEM_USER_NAME = "_system"


def system_user() -> "User":
    """Return the `_system` user."""
    return User.objects.get(username=SYSTEM_USER_NAME)


class UserQuerySet[A](QuerySet["User", A]):
    """Custom QuerySet for User."""

    @permission_filter(work_request=Allow.PASS)
    def can_display(
        self,
        pc: PermissionContext,  # noqa: ARG002
    ) -> "UserQuerySet[A]":
        """Keep only Users that can be displayed."""
        return self

    @permission_filter()
    def can_manage(self, pc: PermissionContext) -> "UserQuerySet[A]":
        """Keep only Users that the given user can manage."""
        assert pc.user.pk is not None  # Enforced by decorator
        return self.filter(pk=pc.user.pk)


class UserManager[A](DjangoUserManager["User"]):
    """Manager for User model."""

    # We cannot use from_queryset or we hit this problem:
    # https://stackoverflow.com/questions/68367703/adding-custom-queryset-to-usermodel-causes-makemigrations-exception  # noqa: E501
    # Therefore we need to proxy all permission filters here

    def can_display(self, pc: PermissionContext) -> "UserQuerySet[A]":
        """Keep only Users that can be displayed."""
        return self.get_queryset().can_display(pc)

    def can_manage(self, pc: PermissionContext) -> "UserQuerySet[A]":
        """Keep only Users that the given user can manage."""
        return self.get_queryset().can_manage(pc)

    @override
    def get_queryset(self) -> UserQuerySet[Any]:
        """Use the custom QuerySet."""
        return UserQuerySet(self.model, using=self._db)


class User(AbstractUser):
    """Debusine user."""

    is_system = models.BooleanField(default=False)

    # mypy seems to struggle changing the manager in a subclass
    objects = UserManager()  # type: ignore[misc]

    class Meta(AbstractUserMeta):
        base_manager_name = "objects"
        constraints = [
            # System users must not have an email address.
            CheckConstraint(
                check=Q(is_system=False) | Q(email=""),
                name="%(app_label)s_%(class)s_non_system_email",
            ),
            # Email addresses of non-system users must be unique.
            UniqueConstraint(
                fields=["email"],
                condition=Q(is_system=False),
                name="%(app_label)s_%(class)s_unique_email",
            ),
            # Enforce non-integer usernames
            CheckConstraint(
                check=~Q(username__regex=r"^\d+$"),
                name="%(app_label)s_%(class)s_non_integer_username",
            ),
        ]

    @permission_check(
        "{user} cannot display user {resource}", work_request=Allow.PASS
    )
    def can_display(self, pc: PermissionContext) -> bool:  # noqa: ARG002, U100
        """Check if the user can be displayed."""
        # Disallow unauthenticated people to enumerate users in the system.
        #
        # In general, we do not assume that user information is automatically
        # public.
        #
        # This could help make life a bit harder, for example, for drive-by
        # spammers trying to enumerate users and their personal information.
        return True

    @permission_check("{user} cannot manage user {resource}")
    def can_manage(self, pc: PermissionContext) -> bool:
        """Check if the user can manage this user."""
        return self.pk == pc.user.pk

    def ui(self, request: "HttpRequest") -> "UserUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.auth import UserUI

        return UserUI.get(request, self)

    def get_absolute_url(self) -> str:
        """Return an absolute URL to this user."""
        return reverse("user:detail", kwargs={"username": self.username})

    def get_absolute_user_id_url(self) -> str:
        """
        Return an absolute numeric URL to this user.

        This is more stable than a username, long-term, for audit purposes.
        """
        return reverse("user:id", kwargs={"pk": self.id})


class Identity(models.Model):
    """
    Identity for a user in a remote user database.

    An Identity is bound if it's associated with a Django user, or unbound if
    no Django user is known for it.
    """

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="identities",
        null=True,
        on_delete=models.SET_NULL,
    )
    issuer = models.CharField(
        max_length=512,
        help_text="identifier of auhoritative system for this identity",
    )
    subject = models.CharField(
        max_length=512,
        help_text="identifier of the user in the issuer system",
    )
    last_used = models.DateTimeField(
        auto_now=True, help_text="last time this identity has been used"
    )
    claims = models.JSONField(default=dict)

    class Meta(TypedModelMeta):
        constraints = [
            UniqueConstraint(
                fields=["issuer", "subject"],
                name="%(app_label)s_%(class)s_unique_issuer_subject",
            ),
        ]

    @override
    def __str__(self) -> str:
        """Return str for the object."""
        return f"{self.issuer}:{self.subject}"


#: Regexp matching the structure of group names
group_name_regex = re.compile(r"^[A-Za-z][A-Za-z0-9+._-]*$")


def is_valid_group_name(value: str) -> bool:
    """Check if value is a valid group name."""
    return bool(group_name_regex.match(value))


def validate_group_name(value: str) -> None:
    """Validate group names."""
    if not is_valid_group_name(value):
        raise ValidationError(
            "%(value)r is not a valid group name", params={"value": value}
        )


class GroupRoleBase(permissions.RoleBase):
    """Group role implementation."""

    implied_by_scope_roles: frozenset["Scope.Roles"]
    implied_by_workspace_roles: frozenset["Workspace.Roles"]
    implied_by_group_roles: frozenset["GroupRoles"]
    implied_by_admin: bool
    implied_by_member: bool

    @override
    def _setup(self) -> None:
        """Set up implications for a newly constructed role."""
        implied_by_scope_roles: set[Scope.Roles] = set()
        implied_by_workspace_roles: set[Workspace.Roles] = set()
        implied_by_group_roles: set[GroupRoles] = {cast(GroupRoles, self)}
        implied_by_admin: bool = False
        implied_by_member: bool = False

        for i in self.implied_by:
            match i:
                case Scope.Roles():
                    implied_by_scope_roles |= i.implied_by_scope_roles
                case Workspace.Roles():
                    implied_by_workspace_roles |= i.implied_by_workspace_roles
                case "member":
                    implied_by_admin = True
                    implied_by_member = True
                case "admin":
                    implied_by_admin = True
                case Role():
                    # Resolve a role passed during class definition into its
                    # enum instance
                    gr = self.__class__(i.value)
                    implied_by_scope_roles |= gr.implied_by_scope_roles
                    implied_by_workspace_roles |= gr.implied_by_workspace_roles
                    implied_by_group_roles |= gr.implied_by_group_roles
                    implied_by_admin = implied_by_admin or gr.implied_by_admin
                    implied_by_member = (
                        implied_by_member or gr.implied_by_member
                    )
                case _:
                    raise ImproperlyConfigured(
                        f"Group roles do not support implications by {type(i)}"
                    )

        self.implied_by_scope_roles = frozenset(implied_by_scope_roles)
        self.implied_by_workspace_roles = frozenset(implied_by_workspace_roles)
        self.implied_by_group_roles = frozenset(implied_by_group_roles)
        self.implied_by_admin = implied_by_admin
        self.implied_by_member = implied_by_member

    def q(self, pc: PermissionContext) -> Q:
        """Return a Q expression to select groups with this role."""
        if not pc.user.is_authenticated:
            return Q(pk__in=())
        q = Q()
        if self.implied_by_member:
            q |= Q(pk__in=pc.groups())

        # Using an assert instead of an if, as implied_by_admin is currently
        # always True
        assert self.implied_by_admin
        q |= Q(membership__user=pc.user, membership__role=Group.Roles.ADMIN)

        for sr in self.implied_by_scope_roles:
            q |= Q(scope__in=Scope.objects.filter(sr.q(pc)))
        for wr in self.implied_by_workspace_roles:
            q |= Q(workspace__isnull=False) & Q(
                workspace__in=Workspace.objects.filter(wr.q(pc))
            )
        return q

    def implies(self, role: "GroupRoles") -> bool:
        """Check if this role implies the given one."""
        return self in role.implied_by_group_roles


class GroupRoles(permissions.Roles, GroupRoleBase, enum.ReprEnum):
    """Available roles for a Scope."""

    ADMIN = Role(
        "admin",
        implied_by=[Scope.Roles.OWNER, Workspace.Roles.OWNER, "admin"],
        description="Add and remove other users",
    )
    MEMBER = Role("member", implied_by=[ADMIN, "member"], description="Member")


GroupRoles.setup()


class GroupQuerySet[A](QuerySet["Group", A]):
    """Custom QuerySet for Group."""

    def with_role(self, pc: PermissionContext, role: GroupRoles) -> Self:
        """Keep only groups where the user has the given role."""
        return self.filter(
            Exists(self.model.objects.filter(role.q(pc), pk=OuterRef("pk")))
        )

    @permission_filter()
    def can_display(
        self,
        pc: PermissionContext,  # noqa: ARG002
    ) -> "GroupQuerySet[A]":
        """Keep only groups that can be displayed."""
        return self.all()

    @permission_filter()
    def can_display_audit_log(
        self, pc: PermissionContext
    ) -> "GroupQuerySet[A]":
        """Keep only groups whose audit log can be displayed."""
        return self.can_display(pc)

    @permission_filter()
    def can_manage(self, pc: PermissionContext) -> "GroupQuerySet[A]":
        """Keep only groups that can be managed."""
        return self.with_role(pc, GroupRoles.ADMIN)

    @permission_filter()
    def can_grant(self, pc: PermissionContext) -> "GroupQuerySet[A]":
        """Keep only groups that can be granted in workflow templates."""
        return self.with_role(pc, GroupRoles.MEMBER)

    def get_by_name(self, name: str) -> "Group":
        """Lookup a group from a scopename/[workspacename/]groupname string."""
        tokens = name.split("/")
        match tokens:
            case [scope, name]:
                res = self.get(scope__name=scope, workspace=None, name=name)
                assert isinstance(res, Group)
                return res
            case [scope, workspace, name]:
                res = self.get(
                    scope__name=scope, workspace__name=workspace, name=name
                )
                assert isinstance(res, Group)
                return res
            case _:
                raise ValueError(
                    f"group name {name!r} should be in the form"
                    " 'scopename[/workspacename]/groupname'"
                )

    def unused_ephemeral(self) -> "GroupQuerySet[Any]":
        """List ephemeral groups with no roles assigned."""
        constraints = Q()
        # Introspect reverse foreign keys to detect which resources have a
        # *Roles table
        for field in Group._meta.get_fields():
            if (
                field.auto_created
                and not field.concrete
                and field.name.endswith("_roles")
            ):
                # If this resource is currently assigning roles, then the group
                # is still in use
                constraints &= Q(**{f"{field.name}__isnull": True})
        return self.filter(ephemeral=True).filter(constraints)


class GroupManager(models.Manager["Group"]):
    """Manager for Group model."""

    @override
    def get_queryset(self) -> GroupQuerySet[Any]:
        """Use the custom QuerySet."""
        return GroupQuerySet(self.model, using=self._db)

    def create_ephemeral(
        self, workspace: Workspace, name: str, owner: User
    ) -> "Group":
        """Create an ephemeral group for the given user."""
        group = self.create(
            scope=workspace.scope,
            workspace=workspace,
            name=name,
            ephemeral=True,
        )
        # Sets the user as an admin of the newly created group
        group.add_user(owner, Group.Roles.ADMIN)
        return group


class Group(models.Model):
    """
    Connect users to roles.

    Group names are scoped: groups with the same name in different scopes are
    entirely distinct and unrelated.
    """

    Roles: TypeAlias = GroupRoles

    name = models.CharField(max_length=255, validators=[validate_group_name])
    scope = models.ForeignKey(Scope, null=False, on_delete=models.PROTECT)
    workspace = models.ForeignKey(
        Workspace, null=True, blank=True, on_delete=models.PROTECT
    )
    users = models.ManyToManyField(
        User,
        blank=True,
        through="db.GroupMembership",
        related_name="debusine_groups",
    )
    ephemeral = models.BooleanField(
        default=False, help_text="remove the group if it has no roles assigned"
    )

    objects = GroupManager.from_queryset(GroupQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"
        constraints = [
            UniqueConstraint(
                F("name"),
                F("scope"),
                # Hack to simulate nulls_distinct in Django 4.2
                Coalesce("workspace", Value(-1)),
                name="%(app_label)s_%(class)s_unique_name_scope_workspace",
            ),
            # From Django 5.0
            # UniqueConstraint(
            #     fields=["name", "scope", "workspace"],
            #     name="%(app_label)s_%(class)s_unique_name_scope_workspace",
            #     nulls_distinct=False,
            # ),
        ]

    @override
    def __str__(self) -> str:
        """Return the scoped group name."""
        if self.workspace is not None:
            return f"{self.scope.name}/{self.workspace.name}/{self.name}"
        else:
            return f"{self.scope.name}/{self.name}"

    @override
    @copy_signature_from(models.Model.save)
    def save(self, *args: Any, **kwargs: Any) -> None:
        """Update audit log on group create/update."""
        old: Group | None = None
        if self.pk:
            old = Group.objects.get(pk=self.pk)
        super().save(*args, **kwargs)
        if old is None:
            self.add_audit_log(GroupAuditLogCreated.create(self))
        else:
            self.add_audit_log(GroupAuditLogUpdated.create_diff(old, self))

    def get_absolute_url(self) -> str:
        """Return an absolute URL to this group."""
        # Prevent circular import
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.scope.name):
            if self.workspace is None:
                return reverse(
                    "groups:detail",
                    kwargs={"group": self.name},
                )
            else:
                return reverse(
                    "workspaces:groups:detail",
                    kwargs={"wname": self.workspace.name, "group": self.name},
                )

    @property
    def name_without_scope(self) -> str:
        """Get the qualified group name without the scope."""
        if self.workspace is not None:
            return f"{self.workspace.name}/{self.name}"
        else:
            return self.name

    def ui(self, request: "HttpRequest") -> "GroupUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.auth import GroupUI

        return GroupUI.get(request, self)

    def get_absolute_url_audit_log(self) -> str:
        """Return an absolute URL to this group's audit log."""
        # Prevent circular import
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.scope.name):
            if self.workspace is None:
                return reverse(
                    "groups:audit-log",
                    kwargs={"group": self.name},
                )
            else:
                return reverse(
                    "workspaces:groups:audit-log",
                    kwargs={"wname": self.workspace.name, "group": self.name},
                )

    @permission_check("{user} cannot display {resource}")
    def can_display(self, pc: PermissionContext) -> bool:  # noqa: ARG002
        """Check if the group can be displayed."""
        return True

    @permission_check("{user} cannot display the audit log of {resource}")
    def can_display_audit_log(self, pc: PermissionContext) -> bool:
        """Check if the group audit log can be displayed."""
        return self.can_display(pc)

    @permission_check("{user} cannot manage {resource}")
    def can_manage(self, pc: PermissionContext) -> bool:
        """Check if the group can be managed."""
        from debusine.db.context import context

        # Shortcut to avoid hitting the database for common cases
        if self.scope == context.scope and context.pc_is(pc):
            if not context.scope_roles.isdisjoint(
                GroupRoles.ADMIN.implied_by_scope_roles
            ):
                return True
            if context.workspace and not context.workspace_roles.isdisjoint(
                GroupRoles.ADMIN.implied_by_workspace_roles
            ):
                return True
        return Group.objects.can_manage(pc).filter(pk=self.pk).exists()

    @permission_check("{user} cannot grant {resource}")
    def can_grant(self, pc: PermissionContext) -> bool:
        """Check if the group can be granted in workflow templates."""
        from debusine.db.context import context

        # Shortcut to avoid hitting the database for common cases
        if self.scope == context.scope and context.pc_is(pc):
            if not context.scope_roles.isdisjoint(
                GroupRoles.ADMIN.implied_by_scope_roles
            ):
                return True
            if context.workspace and not context.workspace_roles.isdisjoint(
                GroupRoles.ADMIN.implied_by_workspace_roles
            ):
                return True
        return Group.objects.can_grant(pc).filter(pk=self.pk).exists()

    def assign_role(self, resource: models.Model, role: str) -> models.Model:
        """Assign a role on a resource."""
        if get_resource_scope(resource) != self.scope:
            raise ValueError(
                f"{resource.__class__.__name__} {str(resource)!r}"
                f" is not in scope {self.scope}"
            )

        role = resolve_role(resource, role)

        roles_model_cls: models.Model = getattr(
            resource.__class__, "objects"
        ).get_roles_model()
        assignment, _ = getattr(roles_model_cls, "objects").get_or_create(
            resource=resource, group=self, role=role
        )
        return cast(models.Model, assignment)

    def get_user_role(self, user: User) -> GroupRoles:
        """Get the role of an existing group member."""
        try:
            membership = GroupMembership.objects.get(group=self, user=user)
        except GroupMembership.DoesNotExist:
            raise ValueError(f"User {user} is not a member of group {self}")
        return cast(GroupRoles, membership.role)

    def set_user_role(self, user: User, role: Roles) -> "GroupMembership":
        """Set the role of an existing group member."""
        try:
            membership = GroupMembership.objects.get(group=self, user=user)
        except GroupMembership.DoesNotExist:
            raise ValueError(f"User {user} is not a member of group {self}")
        membership.role = role
        membership.save()
        self.add_audit_log(
            GroupAuditLogMemberRoleChanged(user=user.username, role=role)
        )
        return membership

    def add_user(
        self, user: User, role: Roles = Roles.MEMBER
    ) -> "GroupMembership":
        """Add a user to the group."""
        membership, created = GroupMembership.objects.get_or_create(
            group=self, user=user, defaults={"role": role}
        )
        if not created:
            raise ValueError(f"{user} is already a member of group {self}")
        self.add_audit_log(
            GroupAuditLogMemberAdded(user=user.username, role=role)
        )
        return membership

    def remove_user(self, user: User) -> None:
        """Remove an existing member from the group."""
        try:
            membership = GroupMembership.objects.get(group=self, user=user)
        except GroupMembership.DoesNotExist:
            raise ValueError(f"User {user} is not a member of group {self}")
        membership.delete()
        self.add_audit_log(GroupAuditLogMemberRemoved(user=user.username))

    def add_audit_log(
        self, changes: "GroupAuditLogEntry"
    ) -> Optional["GroupAuditLog"]:
        """Add an audit log entry."""
        from debusine.db.context import ContextConsistencyError, context

        if context.user is None or not context.user.is_authenticated:
            if context.permission_checks_disabled:
                return None
            raise ContextConsistencyError("add_audit_log requires a valid user")
        changes.actor = context.user.username
        return GroupAuditLog.objects.create_entry(
            group=self, actor=context.user, changes=changes
        )


class GroupMembership(models.Model):
    """Membership of a user in a group."""

    Roles: TypeAlias = GroupRoles

    group = models.ForeignKey(
        Group, on_delete=models.CASCADE, related_name="membership"
    )
    user = models.ForeignKey(
        User, on_delete=models.CASCADE, related_name="group_memberships"
    )
    role = models.CharField(
        max_length=16, choices=Roles.choices, default=Roles.MEMBER
    )

    class Meta(TypedModelMeta):
        constraints = [
            UniqueConstraint(
                fields=["group", "user"],
                name="%(app_label)s_%(class)s_unique_group_user",
            ),
        ]

    @override
    def __str__(self) -> str:
        """Return a string representation."""
        return f"{self.group}:{self.user}"

    def ui(self, request: "HttpRequest") -> "GroupMembershipUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.auth import GroupMembershipUI

        return GroupMembershipUI.get(request, self)

    def get_absolute_url_update(self) -> str:
        """Return an absolute URL to manage this memberhsip."""
        # Prevent circular import
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.group.scope.name):
            if self.group.workspace is None:
                return reverse(
                    "groups:update-member",
                    kwargs={
                        "group": self.group.name,
                        "user": self.user.username,
                    },
                )
            else:
                return reverse(
                    "workspaces:groups:update-member",
                    kwargs={
                        "wname": self.group.workspace.name,
                        "group": self.group.name,
                        "user": self.user.username,
                    },
                )

    def get_absolute_url_remove(self) -> str:
        """Return an absolute URL to remove this memberhsip."""
        # Prevent circular import
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.group.scope.name):
            if self.group.workspace is None:
                return reverse(
                    "groups:remove-member",
                    kwargs={
                        "group": self.group.name,
                        "user": self.user.username,
                    },
                )
            else:
                return reverse(
                    "workspaces:groups:remove-member",
                    kwargs={
                        "wname": self.group.workspace.name,
                        "group": self.group.name,
                        "user": self.user.username,
                    },
                )


class GroupAuditLogEntryTypes(StrEnum):
    """Possible values for GroupAuditLogEntry types."""

    CREATED = "create"
    UPDATED = "update"
    ADD_USER = "add_user"
    REMOVE_USER = "del_user"
    SET_USER_ROLE = "set_role"


class GroupAuditLogEntryBase(pydantic.BaseModel, abc.ABC):
    """Base class for all audit log payloads."""

    model_config = pydantic.ConfigDict(validate_assignment=True, extra="forbid")

    #: Username of the user performing the action
    actor: str = ""

    @abc.abstractmethod
    def description(self) -> str:
        """Format a description for this change."""


class GroupAuditLogCreated(GroupAuditLogEntryBase):
    """Group has been created."""

    type: Literal[GroupAuditLogEntryTypes.CREATED] = (
        GroupAuditLogEntryTypes.CREATED
    )
    name: str
    scope: str
    workspace: str | None = None
    ephemeral: bool

    @classmethod
    def create(cls, group: Group) -> Self:
        """Create an entry populated with group's fields."""
        return cls(
            name=group.name,
            scope=str(group.scope),
            workspace=group.workspace.name if group.workspace else None,
            ephemeral=group.ephemeral,
        )

    @override
    def description(self) -> str:
        unknown = "<unknown>"
        group_name = self.name or unknown
        if self.workspace is not None:
            group_name = self.workspace + "/" + group_name
        group_desc = f"group {group_name} in scope {self.scope or unknown}"
        if self.ephemeral:
            return f"{self.actor} created ephemeral {group_desc}"
        else:
            return f"{self.actor} created {group_desc}"


class GroupAuditLogUpdated(GroupAuditLogEntryBase):
    """Group has been updated."""

    type: Literal[GroupAuditLogEntryTypes.UPDATED] = (
        GroupAuditLogEntryTypes.UPDATED
    )
    name: tuple[str, str] | None = None
    scope: tuple[str, str] | None = None
    workspace: tuple[str | None, str | None] | None = None
    ephemeral: tuple[bool, bool] | None = None

    @classmethod
    def create_diff(cls, old: Group, new: Group) -> Self:
        """
        Compute the changes between two Group objects.

        :returns: a dict mapping field names to tuples of ``(old, new)`` values
        """
        name: tuple[str, str] | None = None
        scope: tuple[str, str] | None = None
        workspace: tuple[str | None, str | None] | None = None
        ephemeral: tuple[bool, bool] | None = None
        if old.name != new.name:
            name = (old.name, new.name)
        if old.scope != new.scope:
            scope = (str(old.scope), str(new.scope))
        if old.workspace != new.workspace:
            workspace = (
                old.workspace.name if old.workspace else None,
                new.workspace.name if new.workspace else None,
            )
        if old.ephemeral != new.ephemeral:
            ephemeral = (old.ephemeral, new.ephemeral)
        return cls(
            name=name, scope=scope, workspace=workspace, ephemeral=ephemeral
        )

    @override
    def description(self) -> str:
        changes: list[str] = []
        for field in "name", "scope", "workspace", "ephemeral":
            change = getattr(self, field)
            if change is None:
                continue
            changes.append(f"{field}: {change[0]!r}→{change[1]!r}")
        if changes:
            return f"{self.actor} updated {'; '.join(changes)}"
        else:
            return f"{self.actor} updated group, no changes detected"


class GroupAuditLogMembershipBase(GroupAuditLogEntryBase):
    """Base for membership changes descriptions."""

    #: Username of the member affected
    user: str


class GroupAuditLogMemberAdded(GroupAuditLogMembershipBase):
    """Member added to the gruoup."""

    type: Literal[GroupAuditLogEntryTypes.ADD_USER] = (
        GroupAuditLogEntryTypes.ADD_USER
    )
    role: GroupRoles

    @override
    def description(self) -> str:
        return f"{self.actor} added {self.user} as {self.role}"


class GroupAuditLogMemberRoleChanged(GroupAuditLogMembershipBase):
    """Member role updated."""

    type: Literal[GroupAuditLogEntryTypes.SET_USER_ROLE] = (
        GroupAuditLogEntryTypes.SET_USER_ROLE
    )
    role: GroupRoles

    @override
    def description(self) -> str:
        return f"{self.actor} set {self.user}'s role as {self.role}"


class GroupAuditLogMemberRemoved(GroupAuditLogMembershipBase):
    """Member removed from the gruoup."""

    type: Literal[GroupAuditLogEntryTypes.REMOVE_USER] = (
        GroupAuditLogEntryTypes.REMOVE_USER
    )

    @override
    def description(self) -> str:
        return f"{self.actor} removed {self.user}"


GroupAuditLogEntry: TypeAlias = Annotated[
    GroupAuditLogCreated
    | GroupAuditLogUpdated
    | GroupAuditLogMemberAdded
    | GroupAuditLogMemberRoleChanged
    | GroupAuditLogMemberRemoved,
    pydantic.Field(discriminator="type"),
]

group_audit_log_entry_adapter: pydantic.TypeAdapter[GroupAuditLogEntry] = (
    pydantic.TypeAdapter(GroupAuditLogEntry)
)


class GroupAuditLogManager(models.Manager["GroupAuditLog"]):
    """Manager for :py:class:`GroupAuditLog`."""

    def create_entry(
        self,
        group: Group,
        actor: User,
        changes: GroupAuditLogEntry,
    ) -> "GroupAuditLog":
        """Create GroupAuditLog entry."""
        return self.create(
            group=group, actor=actor, changes=changes.model_dump()
        )


class GroupAuditLog(models.Model):
    """Audit log entry for group changes."""

    created_at = models.DateTimeField(auto_now_add=True)
    group = models.ForeignKey(
        Group, on_delete=models.CASCADE, related_name="audit_log"
    )
    actor = models.ForeignKey(User, on_delete=models.SET_NULL, null=True)
    changes = models.JSONField()

    objects = GroupAuditLogManager()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"

    @property
    def changes_model(self) -> GroupAuditLogEntry:
        """Return the pydantic model for changes."""
        return group_audit_log_entry_adapter.validate_python(self.changes)


class ClientEnroll(models.Model):
    """Information from a client wanting to enroll."""

    created_at = models.DateTimeField(auto_now_add=True)
    nonce = models.CharField(max_length=64, unique=True)
    payload = models.JSONField(default=dict)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the test utility functions."""

import contextlib
import datetime as dt
import io
from unittest import TestCase

from django.test import TestCase as DjangoTestCase

from debusine.db.models import Scope
from debusine.test import test_utils
from debusine.test.test_utils import (
    create_work_request_response,
    date_time_to_isoformat_rest_framework,
    print_sql_and_stack,
    tomorrow,
    yesterday,
)


class UtilsTests(TestCase):
    """Tests for functions in the utils module."""

    def test_data_generator(self) -> None:
        """Test utils.data_generator: assert type and properties of the data."""
        size = 100

        data_generator = test_utils.data_generator(size)

        data_1 = next(data_generator)

        self.assertIsInstance(data_1, bytes)
        self.assertEqual(len(data_1), size)

        half = int(size / 2)

        # Generated data should not be all the same. Ensure that
        # second half is different to the first half
        self.assertNotEqual(data_1[0:half], data_1[half:size])
        self.assertEqual(len(data_1[0:half]), len(data_1[half:size]))

        data_2 = next(data_generator)

        # Calling it again does not return the same data
        self.assertNotEqual(data_1, data_2)

    def test_yesterday(self) -> None:
        """Test utils.yesterday: return a time older than 24 from now in UTC."""
        yesterday_date = yesterday()

        # time zone aware
        self.assertEqual(yesterday_date.tzinfo, dt.UTC)

        delta = yesterday_date - dt.datetime.now(dt.UTC) - dt.timedelta(days=1)

        self.assertLess(delta, dt.timedelta(seconds=1))

    def test_tomorrow(self) -> None:
        """Test utils.tomorrow: return a time older than 24 from now in UTC."""
        tomorrow_date = tomorrow()

        # time zone aware
        self.assertEqual(tomorrow_date.tzinfo, dt.UTC)

        delta = dt.datetime.now(dt.UTC) + dt.timedelta(days=1) - tomorrow_date

        self.assertLessEqual(delta, dt.timedelta(seconds=1))

    def test_date_time_to_isoformat_rest_framework_utc(self) -> None:
        """Test method when timezone is UTC."""
        value = dt.datetime(2023, 4, 21, 13, 14, 43, 999949, tzinfo=dt.UTC)
        self.assertEqual(
            date_time_to_isoformat_rest_framework(value),
            "2023-04-21T13:14:43.999949Z",
        )

    def test_date_time_to_isoformat_rest_framework_another_timezone(
        self,
    ) -> None:
        """Test method when timezone is not UTC."""
        value = dt.datetime(
            2023,
            4,
            21,
            13,
            14,
            43,
            999949,
            tzinfo=dt.timezone(dt.timedelta(hours=1)),
        )
        self.assertEqual(
            date_time_to_isoformat_rest_framework(value),
            "2023-04-21T13:14:43.999949+01:00",
        )

    def test_create_work_request_response_defaults(self) -> None:
        response = create_work_request_response()
        self.assertNotIn("webui_aborted_by", response.links)
        self.assertIn("webui_created_by", response.links)
        self.assertIn("parent", response.links)
        self.assertIn("workflow_root", response.links)

    def test_create_work_request_response_aborted(self) -> None:
        response = create_work_request_response(aborted_by=12)
        self.assertIn("webui_aborted_by", response.links)

    def test_create_work_request_response_standalone(self) -> None:
        response = create_work_request_response(parent=None, workflow_root=None)
        self.assertNotIn("parent", response.links)
        self.assertNotIn("workflow_root", response.links)


class UtilsDatabaseTests(DjangoTestCase):
    """Tests for functions in the utils module that need a database."""

    def test_print_sql_and_stack(self) -> None:
        """Test print_sql_and_stack."""
        # Nothing is printed if no queries are made
        with (
            contextlib.redirect_stderr(io.StringIO()) as output,
            print_sql_and_stack(),
        ):
            pass
        self.assertEqual(output.getvalue(), "")

        # Output happens with queries
        with (
            contextlib.redirect_stderr(io.StringIO()) as output,
            print_sql_and_stack(),
        ):
            list(Scope.objects.all())
        self.assertEqual(output.getvalue()[:4], "#1: ")
        self.assertIn("test_print_sql_and_stack", output.getvalue())
        self.assertIn("SELECT", output.getvalue())



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the open metrics view."""

from collections.abc import Callable, Iterable
from typing import assert_never, cast, override
from unittest import mock

import requests
from django.test import SimpleTestCase
from django.urls import reverse
from django_prometheus.testutils import assert_metric_diff, save_registry
from prometheus_client import CONTENT_TYPE_LATEST as PROMETHEUS_CONTENT_TYPE
from prometheus_client import (
    CollectorRegistry,
    Counter,
    Gauge,
    Histogram,
    Metric,
    Summary,
)
from prometheus_client.openmetrics.exposition import (
    CONTENT_TYPE_LATEST as OPENMETRICS_CONTENT_TYPE,
)
from prometheus_client.openmetrics.parser import (
    text_string_to_metric_families as openmetrics_parser,
)
from prometheus_client.parser import (
    text_string_to_metric_families as prometheus_parser,
)

from debusine.server.views.open_metrics import (
    RendererFormats,
    extract_media_type,
)
from debusine.tasks.models import WorkerType
from debusine.test.django import TestCase


class HelperTests(SimpleTestCase):
    """Tests for simple helper methods."""

    def test_extract_media_type(self) -> None:
        self.assertEqual(
            extract_media_type("text/plain; version=1; charset=utf-8"),
            "text/plain",
        )


class OpenMetricsViewTests(TestCase):
    """Tests for OpenMetricsView."""

    def get_metrics(
        self, renderer_format: RendererFormats = RendererFormats.PROMETHEUS
    ) -> dict[str, Metric]:
        headers: dict[str, str]
        expected_content_type: str
        parser: Callable[[str], Iterable[Metric]]

        match renderer_format:
            case RendererFormats.PROMETHEUS:
                headers = {}
                expected_content_type = PROMETHEUS_CONTENT_TYPE
                parser = prometheus_parser
            case RendererFormats.OPENMETRICS:
                headers = {"Accept": "application/openmetrics-text"}
                expected_content_type = OPENMETRICS_CONTENT_TYPE
                parser = cast(
                    Callable[[str], Iterable[Metric]], openmetrics_parser
                )
            case _ as unreachable:
                assert_never(unreachable)

        response = self.client.get(reverse("api:open-metrics"), headers=headers)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.headers["Content-Type"], expected_content_type
        )
        text = response.content.decode()
        if renderer_format == RendererFormats.OPENMETRICS:
            self.assertEqual(text.splitlines().count("# EOF"), 1)
        return {metric.name: metric for metric in parser(text)}

    def test_output_headers_prometheus(self) -> None:
        self.get_metrics(RendererFormats.PROMETHEUS)

    def test_output_headers_openmetrics(self) -> None:
        self.get_metrics(RendererFormats.OPENMETRICS)

    def test_default_collector_metrics_disabled(self) -> None:
        metrics = self.get_metrics()

        self.assertNotIn("python_gc_objects_collected", metrics)
        self.assertNotIn("python_info", metrics)
        self.assertNotIn("process_virtual_memory_bytes", metrics)

    def test_default_collector_metrics_disabled_single_process(self) -> None:
        metrics = self.get_metrics()

        self.assertNotIn("python_gc_objects_collected", metrics)
        self.assertNotIn("python_info", metrics)
        self.assertNotIn("process_virtual_memory_bytes", metrics)

    def test_non_database_metrics_registered(self) -> None:
        # Ensure that websocket metrics are registered.
        import debusine.server.consumers  # noqa: F401

        self.client.get(reverse("homepage:homepage"))

        metrics = self.get_metrics()
        self.assertIn(
            "django_http_responses_total_by_status_view_method", metrics
        )
        responses = {
            (
                sample.labels["status"],
                sample.labels["view"],
                sample.labels["method"],
            ): sample.value
            for sample in metrics[
                "django_http_responses_total_by_status_view_method"
            ].samples
        }
        self.assertGreaterEqual(
            responses[("200", "homepage:homepage", "GET")], 1.0
        )
        self.assertIn("debusine_active_websocket_connections", metrics)


class OpenMetricsEmitViewTests(TestCase):
    """Tests for OpenMetricsEmitView."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.registry = CollectorRegistry()
        self.prefix = self.__class__.__name__
        self.counter = Counter(
            f"{self.prefix}_test_counter",
            "test_counter",
            labelnames=("foo",),
            registry=self.registry,
        )
        self.gauge = Gauge(
            f"{self.prefix}_test_gauge",
            "test_gauge",
            labelnames=("foo",),
            registry=self.registry,
        )
        self.summary = Summary(
            f"{self.prefix}_test_summary",
            "test_summary",
            labelnames=("foo",),
            registry=self.registry,
        )
        self.histogram = Histogram(
            f"{self.prefix}_test_histogram",
            "test_histogram",
            labelnames=("foo",),
            registry=self.registry,
        )
        self.enterContext(
            mock.patch("debusine.db.metrics.REGISTRY", new=self.registry)
        )
        self.frozen_registry = save_registry(registry=self.registry)
        self.worker = self.playground.create_worker(
            worker_type=WorkerType.CELERY
        )
        assert self.worker.token is not None
        self.token = self.worker.token

    def test_anonymous(self) -> None:
        response = self.client.post(
            reverse("api:open-metrics-emit"),
            data={
                "metric_type": "counter",
                "name": f"{self.prefix}_test_counter",
                "labels": {},
                "value": 1,
            },
            content_type="application/json",
        )
        self.assertResponseProblem(
            response,
            "Error",
            detail_pattern=r"Authentication credentials were not provided\.",
            status_code=requests.codes.forbidden,
        )

    def test_external_worker_token(self) -> None:
        token = self.playground.create_worker_token()
        response = self.client.post(
            reverse("api:open-metrics-emit"),
            data={
                "metric_type": "counter",
                "name": f"{self.prefix}_test_counter",
                "labels": {},
                "value": 1,
            },
            headers={"token": token.key},
            content_type="application/json",
        )
        self.assertResponseProblem(
            response,
            "Error",
            detail_pattern=r"Endpoint is not accessible to external workers",
            status_code=requests.codes.forbidden,
        )

    def test_user_auth(self) -> None:
        token = self.playground.create_user_token()
        response = self.client.post(
            reverse("api:open-metrics-emit"),
            data={
                "metric_type": "counter",
                "name": f"{self.prefix}_test_counter",
                "labels": {},
                "value": 1,
            },
            headers={"token": token.key},
            content_type="application/json",
        )
        self.assertResponseProblem(
            response,
            "Error",
            detail_pattern=(
                r"You do not have permission to perform this action\."
            ),
            status_code=requests.codes.forbidden,
        )

    def test_metric_not_found(self) -> None:
        response = self.client.post(
            reverse("api:open-metrics-emit"),
            data={
                "metric_type": "counter",
                "name": "nonexistent",
                "labels": {},
                "value": 1,
            },
            headers={"token": self.token.key},
            content_type="application/json",
        )
        self.assertResponseProblem(
            response,
            "Metric not found",
            detail_pattern="Metric 'nonexistent' not found",
            status_code=requests.codes.not_found,
        )

    def test_bad_metric_type(self) -> None:
        for metric_type, name in (
            ("counter", f"{self.prefix}_test_gauge"),
            ("gauge", f"{self.prefix}_test_counter"),
            ("summary", f"{self.prefix}_test_histogram"),
            ("histogram", f"{self.prefix}_test_summary"),
        ):
            with self.subTest(metric_type=metric_type):
                response = self.client.post(
                    reverse("api:open-metrics-emit"),
                    data={
                        "metric_type": metric_type,
                        "name": name,
                        "labels": {},
                        "value": 1,
                    },
                    headers={"token": self.token.key},
                    content_type="application/json",
                )
                self.assertResponseProblem(
                    response,
                    "Bad metric type",
                    detail_pattern=(
                        f"Metric '{name}' is not a {metric_type.capitalize()}"
                    ),
                )

    def test_success(self) -> None:
        for metric_type, name, value, expected_diffs in (
            (
                "counter",
                f"{self.prefix}_test_counter",
                1,
                ((f"{self.prefix}_test_counter_total", 1, {}),),
            ),
            (
                "gauge",
                f"{self.prefix}_test_gauge",
                2.5,
                ((f"{self.prefix}_test_gauge", 2.5, {}),),
            ),
            (
                "summary",
                f"{self.prefix}_test_summary",
                512,
                (
                    (f"{self.prefix}_test_summary_count", 1, {}),
                    (f"{self.prefix}_test_summary_sum", 512, {}),
                ),
            ),
            (
                "histogram",
                f"{self.prefix}_test_histogram",
                10.0,
                (
                    (f"{self.prefix}_test_histogram_bucket", 0, {"le": "7.5"}),
                    (f"{self.prefix}_test_histogram_bucket", 1, {"le": "10.0"}),
                    (f"{self.prefix}_test_histogram_count", 1, {}),
                    (f"{self.prefix}_test_histogram_sum", 10.0, {}),
                ),
            ),
        ):
            with self.subTest(metric_type=metric_type):
                response = self.client.post(
                    reverse("api:open-metrics-emit"),
                    data={
                        "metric_type": metric_type,
                        "name": name,
                        "labels": {"foo": "bar"},
                        "value": value,
                    },
                    headers={"token": self.token.key},
                    content_type="application/json",
                )
                self.assertEqual(
                    response.status_code, requests.codes.no_content
                )
                for (
                    expected_name,
                    expected_diff,
                    expected_labels,
                ) in expected_diffs:
                    with self.subTest(expected_name=expected_name):
                        assert_metric_diff(
                            self.frozen_registry,
                            expected_diff,
                            expected_name,
                            registry=self.registry,
                            foo="bar",
                            **expected_labels,
                        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Input fields for DB-side tasks."""

from typing import Unpack, override

from debusine.artifacts.models import CollectionCategory
from debusine.client.models import LookupChildType
from debusine.db.tasks.models import DBArtifactInfo, DBInputCollectionSingle
from debusine.db.tasks.resolver import DBFieldResolver
from debusine.tasks import (
    TaskConfigError,
    ensure_artifact_categories,
    ensure_collection_category,
)
from debusine.tasks.inputs import FieldResolver, serializers
from debusine.tasks.inputs.inputs import (
    BaseArtifactInput,
    DataFieldInput,
    DataFieldInputKwargs,
)
from debusine.tasks.models import BaseTaskData, LookupSingle


class ArtifactInput(BaseArtifactInput[LookupSingle, DBArtifactInfo]):
    """One single input artifact specified in a task data field."""

    serializer = serializers.PydanticSerializer(DBArtifactInfo)

    @override
    def resolve(self, resolver: FieldResolver) -> DBArtifactInfo:
        assert isinstance(resolver, DBFieldResolver)
        lookup = resolver.get_field(self)
        assert lookup is not None
        try:
            artifact = resolver.work_request.lookup_single(
                lookup=lookup,
                expect_type=LookupChildType.ARTIFACT,
            ).artifact
        except LookupError as e:
            raise TaskConfigError(
                f"Lookup for field {self.field!r} in"
                f" {resolver.work_request.workspace} failed: {e}"
            ) from e

        if self.categories is not None:
            ensure_artifact_categories(
                configuration_key=self.field,
                category=artifact.category,
                expected=self.categories,
            )

        return DBArtifactInfo.from_artifact(
            artifact, lookup=lookup, label=self.field
        )

    @override
    def get_artifact_ids(self, value: DBArtifactInfo) -> list[int]:
        return value.resolved_ids()


class CollectionInput(DataFieldInput[LookupSingle, DBInputCollectionSingle]):
    """One single input collection specified in a task data field."""

    serializer = serializers.PydanticSerializer(DBInputCollectionSingle)

    def __init__(
        self,
        *,
        category: CollectionCategory | None = None,
        **kwargs: Unpack[DataFieldInputKwargs],
    ) -> None:
        """
        Look up artifacts enforcing their type.

        :param field: dot-separated list of lookup names in task data
        :param categories: list of acceptable artifact categories
        """
        super().__init__(**kwargs)
        self.category = category

    @override
    def resolve(self, resolver: FieldResolver) -> DBInputCollectionSingle:
        assert isinstance(resolver, DBFieldResolver)
        lookup = resolver.get_field(self)
        assert lookup is not None
        # Look up the collection
        try:
            result = resolver.work_request.lookup_single(
                lookup=lookup,
                default_category=self.category,
                expect_type=LookupChildType.COLLECTION,
            )
        except LookupError as e:
            msg = (
                f"Lookup for field {self.field!r} in"
                f" {resolver.work_request.workspace}"
            )
            if self.category is not None:
                msg += f" with category '{self.category}'"
            msg += f" failed: {e}"
            raise TaskConfigError(msg) from e

        if (
            self.category is not None
            and result.collection.category != self.category
        ):
            ensure_collection_category(
                configuration_key=self.field,
                category=result.collection.category,
                expected=self.category,
            )

        return DBInputCollectionSingle.from_collection(
            result.collection, lookup=lookup, label=self.field
        )


class CollectionInSameWorkspace(CollectionInput):
    """A collection in the same workspace as the work request."""

    @override
    def resolve(self, resolver: FieldResolver) -> DBInputCollectionSingle:
        assert isinstance(resolver, DBFieldResolver)
        info = super().resolve(resolver)
        if info.collection.workspace != resolver.work_request.workspace:
            raise TaskConfigError(
                f"{info.lookup} is in workspace {info.collection.workspace},"
                f" not {resolver.work_request.workspace}"
            )
        return info


class TaskConfigurationCollection(
    DataFieldInput[LookupSingle, DBInputCollectionSingle | None]
):
    """Lookup the task configuration collection to use."""

    serializer = serializers.OptionalPydanticSerializer(DBInputCollectionSingle)

    @override
    def resolve(
        self, resolver: FieldResolver
    ) -> DBInputCollectionSingle | None:
        assert isinstance(resolver, DBFieldResolver)
        if (lookup := resolver.get_field(self)) is None:
            return None

        # Look up the debusine:task-configuration collection
        try:
            lookup_result = resolver.work_request.lookup_single(
                lookup=lookup,
                default_category=CollectionCategory.TASK_CONFIGURATION,
                expect_type=LookupChildType.COLLECTION,
            )
        except LookupError as e:
            if lookup != BaseTaskData.DEFAULT_TASK_CONFIGURATION_LOOKUP:
                raise TaskConfigError(
                    f"Lookup for field {self.field!r} in"
                    f" {resolver.work_request.workspace} failed: {e}"
                ) from e
            else:
                return None

        return DBInputCollectionSingle.from_collection(
            lookup_result.collection, lookup=lookup, label=self.field
        )
